######################################################################################################################
# Copyright (C) 2017-2021 Spine project consortium
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""
Parameter indexing settings widget for .gdx export.

:author: A. Soininen (VTT)
:date:   26.11.2019
"""
from contextlib import contextmanager
import enum
from PySide2.QtCore import Slot
from PySide2.QtWidgets import QWidget
from spinedb_api.spine_io.exporters import gdx
from ..mvcmodels.indexing_table_model import IndexingTableModel


class IndexSettingsState(enum.Enum):
    """An enumeration indicating the state of the settings window."""

    OK = enum.auto()
    DOMAIN_MISSING_INDEXES = enum.auto()


class ParameterIndexSettings(QWidget):
    """A widget showing setting for a parameter with indexed values."""

    def __init__(self, parameter_name, parameter, indexing_setting, available_domains, parent):
        """
        Args:
            parameter_name (str): parameter's name
            parameter (Parameter): parameter to index
            indexing_setting (IndexingSetting): indexing settings for the parameter
            available_domains (dict): a dict from existing domain name to :class:`Records`
            parent (QWidget, optional): a parent widget
        """
        from ..ui.parameter_index_settings import Ui_Form  # pylint: disable=import-outside-toplevel

        super().__init__(parent)
        self._parameter = parameter
        self._indexing_setting = indexing_setting
        self._state = IndexSettingsState.OK
        self._monitor_domains_combo_box = True
        self._using_pick_expression = False
        self._ui = Ui_Form()
        self._ui.setupUi(self)
        self._ui.box.setTitle(parameter_name)
        self._indexing_table_model = IndexingTableModel(self._parameter)
        self._ui.index_table_view.setModel(self._indexing_table_model)
        self._available_domains = available_domains
        self._ui.domains_combo.addItems(sorted(available_domains.keys()))
        self._ui.domains_combo.currentTextChanged.connect(self._change_domain)
        self._ui.pick_expression_edit.textChanged.connect(self._update_index_list_selection)
        self._ui.move_domain_left_button.clicked.connect(self._move_indexing_domain_left)
        self._ui.move_domain_right_button.clicked.connect(self._move_indexing_domain_right)
        indexing_domain_name = indexing_setting.indexing_domain_name
        if indexing_domain_name is None and available_domains:
            indexing_domain_name = next(iter(available_domains))
        if indexing_domain_name is not None:
            if self._ui.domains_combo.currentText() != indexing_domain_name:
                self._ui.domains_combo.setCurrentText(indexing_domain_name)
            else:
                self._change_domain(indexing_domain_name)
            picking = indexing_setting.picking
            if isinstance(picking, gdx.GeneratedPicking):
                self._ui.pick_expression_edit.setText(picking.expression)
            else:
                self._indexing_table_model.set_picking(picking)
        self._check_state()
        self._indexing_table_model.selection_changed.connect(self._check_state)
        self._indexing_table_model.manual_selection.connect(self._clear_pick_expression_silently)

    @property
    def state(self):
        """widget's state"""
        return self._state

    @state.setter
    def state(self, new_state):
        """Sets the state of the widget and possibly shows an error indicator."""
        self._state = new_state
        if self._state == IndexSettingsState.DOMAIN_MISSING_INDEXES:
            self.error_message("Not enough selected indexes to index all values.")
        elif self._state == IndexSettingsState.OK:
            self.notification_message("Parameter successfully indexed.")

    def indexing_domain_name(self):
        """
        Returns the selected indexing domain's name

        Returns:
            str: domain name
        """
        return self._ui.domains_combo.currentText()

    def picking(self):
        """
        Returns picking.

        Returns:
            Picking: picking
        """
        if self._using_pick_expression:
            return gdx.GeneratedPicking(self._ui.pick_expression_edit.text())
        return self._indexing_table_model.get_picking()

    def set_domains_combo_monitoring_enabled(self, enabled):
        """
        Enables or disables monitoring of current text in domains combo box.

        Args:
            enabled (bool): True enables monitoring, False disables
        """
        self._monitor_domains_combo_box = enabled

    def add_domain(self, domain_name):
        """
        Adds a domain to the domains combo box.

        Args:
            domain_name (str) new domain name
        """
        self._ui.domains_combo.addItem(domain_name)

    def remove_domain(self, domain_name):
        """
        Removes a domain from the domains combo box.

        Args:
            domain_name (str): domain to remove
        """
        index = self._ui.domains_combo.findText(domain_name)
        self._ui.domains_combo.removeItem(index)

    def update_domain_name(self, old_name, new_name):
        """
        Renames a domain.

        Args:
            old_name (str): previous name
            new_name (str): new name
        """
        index = self._ui.domains_combo.findText(old_name)
        self._ui.domains_combo.setItemText(index, new_name)

    def update_records(self, domain_name):
        """
        Updates existing domain's records.

        Args:
            domain_name (str): domain's name
        """
        if domain_name == self._ui.domains_combo.currentText():
            self._indexing_table_model.set_records(self._available_domains[domain_name])

    def notification_message(self, message):
        """Shows a notification message on the widget."""
        self._ui.message_label.setText(message)

    def warning_message(self, message):
        """Shows a warning message on the widget."""
        yellow_message = "<span style='color:#b89e00;white-space: pre-wrap;'>" + message + "</span>"
        self._ui.message_label.setText(yellow_message)

    def error_message(self, message):
        """Shows an error message on the widget."""
        red_message = "<span style='color:#ff3333;white-space: pre-wrap;'>" + message + "</span>"
        self._ui.message_label.setText(red_message)

    @Slot()
    def _check_state(self):
        """Updated the widget's state."""
        mapped_values_balance = self._indexing_table_model.mapped_values_balance()
        if self._check_errors(mapped_values_balance):
            return
        if self._check_warnings(mapped_values_balance):
            return
        self.state = IndexSettingsState.OK

    def _check_errors(self, mapped_values_balance):
        """Checks if the parameter is correctly indexed."""
        if mapped_values_balance < 0:
            self.state = IndexSettingsState.DOMAIN_MISSING_INDEXES
            return True
        return False

    def _check_warnings(self, mapped_values_balance):
        """Checks if there are non-fatal issues with parameter indexing."""
        if mapped_values_balance > 0:
            self._state = IndexSettingsState.OK
            self.warning_message("Too many indexes selected. The excess indexes will not be used.")
            return True
        return False

    def _update_indexing_domains_name(self):
        """Updates the model's header and the label showing the indexing domains."""
        index_position = self._indexing_setting.index_position
        domain_name = self._ui.domains_combo.currentText()
        self._indexing_table_model.set_index_name(domain_name)
        name = "<b>{}</b>".format(domain_name if domain_name else "unnamed")
        label = (
            "("
            + ", ".join(
                self._parameter.domain_names[:index_position] + (name,) + self._parameter.domain_names[index_position:]
            )
            + ")"
        )
        self._ui.indexing_domains_label.setText(label)

    @Slot()
    def _clear_pick_expression_silently(self):
        """Clears the pick expression line edit."""
        self._ui.pick_expression_edit.textChanged.disconnect(self._update_index_list_selection)
        self._ui.pick_expression_edit.clear()
        self._using_pick_expression = False
        self._ui.pick_expression_edit.textChanged.connect(self._update_index_list_selection)

    @Slot(str)
    def _change_domain(self, domain_name):
        """Change the domain used on the table."""
        if not self._monitor_domains_combo_box:
            return
        if domain_name:
            self._indexing_table_model.set_records(self._available_domains[domain_name])
        else:
            self._indexing_table_model.set_records(gdx.LiteralRecords([]))
        self._update_indexing_domains_name()

    @Slot(str)
    def _update_index_list_selection(self, expression):
        """Updates selection according to changed selection expression."""
        if not expression:
            self._indexing_table_model.select_all()
            self._using_pick_expression = False
            return
        self._indexing_table_model.set_picking(gdx.GeneratedPicking(expression))
        self._using_pick_expression = True

    @Slot(bool)
    def _move_indexing_domain_left(self, _):
        """Moves the indexing domain name left on the indexing label."""
        if self._indexing_setting.index_position > 0:
            self._indexing_setting.index_position -= 1
            self._update_indexing_domains_name()

    @Slot(bool)
    def _move_indexing_domain_right(self, _):
        """Moves the indexing domain name right on the indexing label."""
        if self._indexing_setting.index_position < len(self._indexing_setting.domain_names):
            self._indexing_setting.index_position += 1
            self._update_indexing_domains_name()


@contextmanager
def _freely_update_domains_combo(widget):
    """
    A context manager which temporarily disables the monitoring of current text changes in domains combo box.

    Args:
        widget (ParameterIndexSettings): settings widget
    """
    widget.set_domains_combo_monitoring_enabled(False)
    try:
        yield None
    finally:
        widget.set_domains_combo_monitoring_enabled(True)



######################################################################################################################
# Copyright (C) 2017-2021 Spine project consortium
# This file is part of Spine Toolbox.
# Spine Toolbox is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""
Contains a model to handle source tables and import mapping.

:author: A. Soininen (VTT)
:date:   7.10.2021
"""
from dataclasses import dataclass, field
from enum import IntEnum, unique
import re

from PySide2.QtCore import QAbstractItemModel, QModelIndex, Qt, Signal

from spinetoolbox.helpers import unique_name
from spinedb_api.parameter_value import join_value_and_type, split_value_and_type
from spinedb_api import from_database, ParameterValueFormatError
from spinedb_api.import_mapping.import_mapping import ImportMapping, ScenarioBeforeAlternativeMapping
from spinedb_api.import_mapping.import_mapping_compat import (
    parse_named_mapping_spec,
    import_mapping_from_dict,
    parameter_mapping_from_dict,
    parameter_default_value_mapping_from_dict,
    parameter_value_mapping_from_dict,
    unparse_named_mapping_spec,
)
from spinedb_api.import_mapping.type_conversion import FloatConvertSpec, StringConvertSpec, DateTimeConvertSpec
from spinedb_api.mapping import Position
from spinetoolbox.mvcmodels.shared import PARSED_ROLE
from ..commands import (
    RenameMapping,
    SetTableChecked,
    UpdateTableItem,
    SetMappingPositionType,
    SetMappingPosition,
    SetFilterRe,
)
from ..flattened_mappings import FlattenedMappings, VALUE_TYPES
from ..mapping_colors import ERROR_COLOR
from .mappings_model_roles import Role


UNNAMED_TABLE_NAME = "<unnamed table>"


@unique
class FlattenedColumn(IntEnum):
    NAME = 0
    POSITION_TYPE = 1
    POSITION = 2
    REGEXP = 3


@dataclass()
class SourceTableItem:
    """A list item for source tables."""

    name: str
    checked: bool
    checkable: bool = True
    editable: bool = False
    real: bool = True
    mapping_list: list = field(init=False, default_factory=list)

    def append_to_mapping_list(self, list_item):
        """Appends an item to mapping list.

        Args:
            list_item (MappingListItem): item
        """
        list_item.source_table_item = self
        self.mapping_list.append(list_item)

    def insert_to_mapping_list(self, row, list_item):
        """Inserts an item to mapping list.

        Args:
            row (int): insertion position
            list_item (MappingListItem): item
        """
        list_item.source_table_item = self
        self.mapping_list = self.mapping_list[:row] + [list_item] + self.mapping_list[row:]


_dummy = object()


@dataclass()
class MappingListItem:
    """An item in mappings list."""

    name: str
    flattened_mappings: FlattenedMappings = field(init=False)
    source_table_item: SourceTableItem = field(init=False)

    def set_flattened_mappings(self, flattened_mappings):
        """Sets flattened mappings for this item.

        Args:
            flattened_mappings (FlattenedMappings): mappings
        """
        flattened_mappings.mapping_list_item = self
        self.flattened_mappings = flattened_mappings


class MappingsModel(QAbstractItemModel):
    """A tree model that manages source tables and their mappings."""

    msg_error = Signal(str)
    """Emitted when an error message should be displayed."""
    row_or_column_type_recommended = Signal(int, object, object)
    """Emitted when a change in mapping prompts for change in column or row type."""
    multi_column_type_recommended = Signal(object, object)
    """Emitted when all but given columns should be of given type."""

    def __init__(self, undo_stack, parent):
        """
        Args:
            parent (QObject): parent object
        """
        super().__init__(parent)
        self._undo_stack = undo_stack
        self._mappings = [self._make_select_all_tables_item()]

    def real_table_names(self):
        """Returns real table names.

        Returns:
            list of str: real table names
        """
        return [i.name for i in self._mappings if i.real]

    def checked_table_names(self):
        """Returns checked table names.

        Returns:
            list of str: checked table names
        """
        return [i.name for i in self._mappings[1:] if i.checked]

    @staticmethod
    def _make_select_all_tables_item():
        """Creates the first source table list item.

        Returns:
            SourceTableItem: 'select all' item
        """
        return SourceTableItem("Select All", checked=False, real=False)

    def columnCount(self, parent=QModelIndex()):
        if not parent.isValid():
            return 1
        parent_item = parent.internalPointer()
        if isinstance(parent_item, SourceTableItem):
            return 1
        if isinstance(parent_item, MappingListItem):
            return 4
        return 0

    def data(self, index, role=Qt.DisplayRole):
        index_item = index.internalPointer()
        if isinstance(index_item, SourceTableItem):
            return self._table_list_data(index, role)
        if isinstance(index_item, MappingListItem):
            return self._mapping_list_data(index_item, index, role)
        if isinstance(index_item, FlattenedMappings):
            return self._mapping_data(index_item, index, role)
        return None

    def _table_list_data(self, index, role):
        """Returns source table data for given role.

        Args:
            index (QModelIndex): table index
            role (int): Qt's data role

        Returns:
            Any: data
        """
        if role in (Qt.DisplayRole, Qt.EditRole):
            return self._mappings[index.row()].name
        if role == Qt.CheckStateRole:
            return Qt.Checked if self._mappings[index.row()].checked else Qt.Unchecked
        if role == Role.ITEM:
            return self._mappings[index.row()]
        return None

    @staticmethod
    def _mapping_list_data(list_item, index, role):
        """Returns mapping list data for given role.

        Args:
            list_item (MappingListItem): list item
            index (QModelIndex): list index
            role (int): Qt's data role

        Returns:
            Any: data
        """
        if role in (Qt.DisplayRole, Qt.EditRole):
            return list_item.name
        if role == Role.ITEM:
            return list_item
        if role == Role.FLATTENED_MAPPINGS:
            return list_item.flattened_mappings
        return None

    @staticmethod
    def _mapping_data(flattened_mappings, index, role):
        """Returns mapping component data for given role.

        Args:
            flattened_mappings (FlattenedMappings): flattened mappings
            index (QModelIndex): component index
            role (int): Qt's data role

        Returns:
            Any: data
        """
        if role == PARSED_ROLE:
            # Only used for the ParameterValueEditor.
            # At this point, the delegate has already checked that it's a constant parameter (default) value mapping
            m = flattened_mappings.component_at(index.row())
            try:
                return from_database(*split_value_and_type(m.value))
            except ParameterValueFormatError:
                return None
        column = index.column()
        if role in (Qt.DisplayRole, Qt.EditRole):
            if column == FlattenedColumn.NAME:
                return flattened_mappings.display_names[index.row()]
            if column == FlattenedColumn.POSITION_TYPE:
                return flattened_mappings.display_position_type(index.row())
            if column == FlattenedColumn.POSITION:
                display = flattened_mappings.display_position(index.row())
                if display == "<table name>":
                    return flattened_mappings.mapping_list_item.source_table_item.name
                return display
            if column == FlattenedColumn.REGEXP:
                return flattened_mappings.component_at(index.row()).filter_re
            raise RuntimeError("Column out of bounds.")
        if role == Qt.BackgroundRole:
            if column == FlattenedColumn.NAME:
                return flattened_mappings.display_colors[index.row()]
            if column == FlattenedColumn.POSITION:
                issues = flattened_mappings.display_row_issues(index.row())
                if issues:
                    return ERROR_COLOR
                return None
        if role == Qt.ToolTipRole:
            if column == FlattenedColumn.POSITION:
                issues = flattened_mappings.display_row_issues(index.row())
                if issues:
                    return issues
                return None
            if column == FlattenedColumn.REGEXP:
                return "Enter regular expression to filter importer data."
        return None

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        if role == Qt.DisplayRole and orientation == Qt.Horizontal:
            return ("Target", "Source type", "Source ref.", "Filter")[section]
        return None

    def flags(self, index):
        index_item = index.internalPointer()
        if isinstance(index_item, SourceTableItem):
            return self._table_list_flags(index)
        if isinstance(index_item, MappingListItem):
            return self._mapping_list_flags()
        if isinstance(index_item, FlattenedMappings):
            return self._mapping_flags(index_item, index)
        return super().flags(index)

    def _table_list_flags(self, index):
        """Returns flags for source table list.

        Args:
            index (QModelIndex): index

        Returns:
            int: flags
        """
        flags = Qt.ItemIsEnabled | Qt.ItemIsSelectable | Qt.ItemIsUserCheckable
        table_item = self._mappings[index.row()]
        if table_item.checkable:
            flags |= Qt.ItemIsUserCheckable
        if table_item.editable:
            flags |= Qt.ItemIsEditable
        return flags

    @staticmethod
    def _mapping_list_flags():
        """Returns flags for mapping list data.

        Returns:
            int: flags
        """
        return Qt.ItemIsSelectable | Qt.ItemIsEnabled | Qt.ItemIsEditable

    @staticmethod
    def _mapping_flags(flattened_item, index):
        """Returns flags for mapping components.

        Args:
            flattened_item (FlattenedMappings): flattened mappings
            index (QModelIndex): index

        Returns:
            int: flags
        """
        non_editable = Qt.ItemIsEnabled | Qt.ItemIsSelectable
        editable = non_editable | Qt.ItemIsEditable
        column = index.column()
        if column == FlattenedColumn.NAME:
            return non_editable
        if flattened_item.root_mapping.is_pivoted():
            # special case where we have pivoted data
            if index.row() == len(flattened_item.display_names) - 1:
                return non_editable
        if column == FlattenedColumn.POSITION:
            component = flattened_item.component_at(index.row())
            if component.value is None and component.position == Position.header:
                return non_editable
        return editable

    def index(self, row, column, parent=QModelIndex()):
        if row < 0 or column < 0:
            return QModelIndex()
        if not parent.isValid():
            return self.createIndex(row, 0, self._mappings[row])
        parent_item = parent.internalPointer()
        if isinstance(parent_item, SourceTableItem):
            return self.createIndex(row, 0, parent_item.mapping_list[row])
        if isinstance(parent_item, MappingListItem):
            return self.createIndex(row, column, parent_item.flattened_mappings)
        raise RuntimeError("Cannot create index.")

    @staticmethod
    def is_table_index(index):
        """Checks if given index is an index to source table list.

        Args:
            index (QModelIndex): index

        Returns:
            bool: True if index is table index, False otherwise
        """
        return isinstance(index.internalPointer(), SourceTableItem)

    @staticmethod
    def is_mapping_list_index(index):
        """Checks if given index is an index to mapping list.

        Args:
            index (QModelIndex): index

        Returns:
            bool: True if index is mapping list index, False otherwise
        """
        return isinstance(index.internalPointer(), MappingListItem)

    def dummy_parent(self):
        """Creates a dummy parent index that always returns an empty table.

        Returns:
            QModelIndex: dummy index
        """
        return self.createIndex(0, 0, _dummy)

    def parent(self, index):
        index_item = index.internalPointer()
        if isinstance(index_item, SourceTableItem):
            return QModelIndex()
        if isinstance(index_item, MappingListItem):
            for table_row, table_item in enumerate(self._mappings):
                if index_item.source_table_item is table_item:
                    return self.createIndex(table_row, 0, table_item)
        elif isinstance(index_item, FlattenedMappings):
            for list_row, list_item in enumerate(index_item.mapping_list_item.source_table_item.mapping_list):
                if index_item is list_item.flattened_mappings:
                    return self.createIndex(list_row, 0, list_item)
        elif index_item is _dummy:
            return QModelIndex()
        raise RuntimeError("Cannot create parent index.")

    def store(self):
        """Stores source tables, mappings lists and mappings to a dict.

        Returns:
            dict: stored data
        """
        table_dicts = dict()
        for table_item in self._mappings[1:]:
            if not table_item.real:
                continue
            table_dicts[table_item.name] = [
                unparse_named_mapping_spec(list_item.name, list_item.flattened_mappings.root_mapping)
                for list_item in table_item.mapping_list
            ]
        selected_tables = self.checked_table_names()
        return {"table_mappings": table_dicts, "selected_tables": selected_tables}

    def restore(self, mappings_dict):
        """Restores source tables, mapping lists and mappings.

        Args:
            mappings_dict (dict): serialized data
        """
        checked_tables = mappings_dict.get("selected_tables")
        self.beginResetModel()
        self._mappings.clear()
        self._mappings.append(self._make_select_all_tables_item())
        try:
            for table_name, mapping_list_dicts in mappings_dict.get("table_mappings", {}).items():
                checked = table_name in checked_tables if checked_tables is not None else True
                table_item = SourceTableItem(table_name, checked)
                for list_dict in mapping_list_dicts:
                    mapping_name, root_mapping = parse_named_mapping_spec(list_dict)
                    flattened_mappings = FlattenedMappings(root_mapping)
                    list_item = MappingListItem(mapping_name)
                    list_item.set_flattened_mappings(flattened_mappings)
                    table_item.append_to_mapping_list(list_item)
                self._mappings.append(table_item)
        except ValueError as error:
            self.msg_error.emit(f"{error}")
            return
        self.endResetModel()
        self._update_all_checked()

    def rowCount(self, parent=QModelIndex()):
        if not parent.isValid():
            return len(self._mappings)
        parent_item = parent.internalPointer()
        if isinstance(parent_item, SourceTableItem):
            return len(parent_item.mapping_list)
        if isinstance(parent_item, MappingListItem):
            if parent_item.flattened_mappings is None:
                return 0
            return len(parent_item.flattened_mappings.display_names)
        return 0

    def setData(self, index, value, role=Qt.EditRole):
        index_item = index.internalPointer()
        if isinstance(index_item, SourceTableItem):
            return self._set_table_list_data(index_item, index, value, role)
        if isinstance(index_item, MappingListItem):
            return self._set_mapping_list_data(index_item, index, value, role)
        if isinstance(index_item, FlattenedMappings):
            return self._set_mapping_data(index_item, index, value, role)
        raise RuntimeError("Unknown parent type.")

    def _set_table_list_data(self, table_item, index, value, role):
        """Sets data for source table list.

        Args:
            table_item (SourceTableItem): table item
            index (QModelIndex): index
            value (Any): value to set
            role (int): Qt's data role

        Returns:
            bool: True if data was set successfully, False otherwise
        """
        row = index.row()
        if role == Qt.CheckStateRole and table_item.checkable:
            checked = value == Qt.Checked
            if row == 0:
                self._set_multiple_checked_undoable(checked, *range(1, len(self._mappings)))
            else:
                self._undo_stack.push(SetTableChecked(table_item.name, self, checked, row))
            return True
        if role == Qt.EditRole:
            if value == table_item.name:
                return False
            if value in {table.name for table in self._mappings if table.real}:
                self.msg_error.emit(f"There's already a table called {value}.")
                return False
            previous = {"name": table_item.name}
            updates = {"name": value}
            self._undo_stack.push(UpdateTableItem(self, row, previous, updates, row == len(self._mappings) - 1))
            return True
        return False

    def _set_mapping_list_data(self, list_item, index, value, role):
        """Sets data for mapping list.

        Args:
            list_item (MappingListItem): list item
            index (QModelIndex): index
            value (Any): value to set
            role (int): Qt's data role

        Returns:
            bool: True if data was set successfully, False otherwise
        """
        if role == Qt.EditRole:
            if not value or value in {m.name for m in list_item.source_table_item.mapping_list}:
                return False
            previous = list_item.name
            parent_index = index.parent()
            self._undo_stack.push(RenameMapping(parent_index.row(), index.row(), self, value, previous))
            return True
        return False

    def update_table_item(self, row, data_dict, add_empty_row=False, remove_empty_row=False):
        """Updates table item.

         This only happens in file-less mode, when the user is creating tables.

        Args:
            row (int): item row
            data_dict (dict): table item data to update
            add_empty_row (bool): whether or not to add an empty row
            remove_empty_row (bool): whether or not to remove the empty row
        """
        table_item = self._mappings[row]
        for field_name, value in data_dict.items():
            setattr(table_item, field_name, value)
        index = self.index(row, 0)
        if add_empty_row:
            table_item.real = True
            table_item.checkable = True
            table_item.checked = True
            default_flattened_mappings = FlattenedMappings(self._create_default_mapping())
            mapping_list_item = MappingListItem(self._unique_mapping_name(table_item))
            mapping_list_item.set_flattened_mappings(default_flattened_mappings)
            self.beginInsertRows(index, 0, 0)
            table_item.append_to_mapping_list(mapping_list_item)
            self.endInsertRows()
            self.add_empty_row()
        if remove_empty_row:
            table_item.real = False
            table_item.checkable = False
            table_item.checked = False
            self.beginRemoveRows(index, 0, len(table_item.mapping_list) - 1)
            table_item.mapping_list.clear()
            self.endRemoveRows()
            self._remove_last_source_table()
        self.dataChanged.emit(index, index)

    def toggle_checked_tables(self, table_indexes):
        """Toggles the checked status of source tables.

        Args:
            table_indexes (Iterable of QModelIndex): indexes
        """
        table_items = ((i.row(), self._mappings[i.row()]) for i in table_indexes)
        checkables = [i for i in table_items if i[1].checkable and i[1].real]
        all_checked = all(i[1].checked for i in checkables)
        self._set_multiple_checked_undoable(not all_checked, *(i[0] for i in checkables))

    def _set_multiple_checked_undoable(self, checked, *rows):
        """
        Sets the checked status of multiple list items.

        This action is undoable.

        Args:
            checked (bool): True for checked, False for unchecked
            *rows: item rows
        """
        rows = [row for row in rows if self._mappings[row].checked != checked]
        self._undo_stack.push(SetTableChecked(None, self, checked, *rows))

    def set_table_checked(self, checked, *rows):
        """
        Sets the checked status of a source table item.

        Args:
            checked (bool): True for checked, False for unchecked
            *rows: item rows
        """
        min_row = None
        max_row = None
        for row in rows:
            if self._mappings[row].checked == checked:
                continue
            if min_row is None:
                min_row = row
                max_row = row
            else:
                min_row = min(row, min_row)
                max_row = max(row, max_row)
            self._mappings[row].checked = checked
        if min_row is None:
            return
        top_left = self.index(min_row, 0)
        bottom_right = self.index(max_row, 0)
        self.dataChanged.emit(top_left, bottom_right, [Qt.CheckStateRole])
        self._update_all_checked()

    def _update_all_checked(self):
        """Updates the checked state of 'Select All' table item if needed."""
        checkables = (m for m in self._mappings[1:] if m.checkable)
        all_checked = all(m.checked for m in checkables)
        all_checked_item = self._mappings[0]
        if all_checked_item.checked != all_checked:
            self._mappings[0].checked = all_checked
            index = self.index(0, 0)
            self.dataChanged.emit(index, index, [Qt.CheckStateRole])

    def append_new_table_with_mapping(self, table_name, root_mapping):
        """Appends a new source table and adds given ``root_mapping`` as its first mapping.

        Args:
            table_name (str): table name
            root_mapping (ImportMapping, optional): initial mapping; if None,
                an object class mapping will be created
        """
        if root_mapping is None:
            root_mapping = self._create_default_mapping()
        flattened_mappings = FlattenedMappings(root_mapping)
        list_item = MappingListItem("Mapping 1")
        list_item.set_flattened_mappings(flattened_mappings)
        table_item = SourceTableItem(table_name, checked=True)
        table_item.append_to_mapping_list(list_item)
        self.beginInsertRows(QModelIndex(), len(self._mappings), len(self._mappings))
        self._mappings.append(table_item)
        self.endInsertRows()

    def _remove_last_source_table(self):
        """Removes the last source table."""
        self.beginRemoveRows(QModelIndex(), self.rowCount() - 1, self.rowCount() - 1)
        self._mappings.pop(-1)
        self.endRemoveRows()

    def rename_mapping(self, table_row, list_row, name):
        """
        Renames a mapping.

        Args:
            table_row (int): source table row
            list_row (int): mapping list row
            name (str): new name
        """
        table_item = self._mappings[table_row]
        list_item = table_item.mapping_list[list_row]
        if name == list_item.name:
            return
        list_item.name = ""
        if self.has_mapping_name(table_row, name):
            match = re.search(r" [0-9]+$", name)
            prefix = name[: match.start()] if match is not None else name
            name = self._unique_mapping_name(table_item, prefix)
        list_item.name = name
        parent = self.index(table_row, 0)
        index = self.index(list_row, 0, parent)
        self.dataChanged.emit(index, index, [Qt.DisplayRole])

    def set_root_mapping(self, table_row, list_row, root_mapping):
        """
        Sets new root mapping.

        Args:
            table_row (int): source table row index
            list_row (int): mapping list row index
            root_mapping (ImportMapping): root mapping
        """
        table_index = self.index(table_row, 0)
        list_index = self.index(list_row, 0, table_index)
        mapping_list = self._mappings[table_row].mapping_list[list_row]
        flattened_mappings = mapping_list.flattened_mappings
        self.beginRemoveRows(list_index, 0, len(flattened_mappings.display_names) - 1)
        mapping_list.flattened_mappings = None
        self.endRemoveRows()
        flattened_mappings.set_root_mapping(root_mapping)
        self.beginInsertRows(list_index, 0, len(flattened_mappings.display_names) - 1)
        mapping_list.flattened_mappings = flattened_mappings
        self.endInsertRows()
        self.dataChanged.emit(list_index, list_index, [Role.FLATTENED_MAPPINGS])

    def set_mappings_type(self, table_row, list_row, new_type):
        """
        Sets the type for mappings.

        Args:
            table_row (int): source table row index
            list_row (int): mapping list row index
            new_type (str): name of the type
        """
        map_type = {
            "Object class": "ObjectClass",
            "Relationship class": "RelationshipClass",
            "Object group": "ObjectGroup",
            "Alternative": "Alternative",
            "Scenario": "Scenario",
            "Scenario alternative": "ScenarioAlternative",
            "Parameter value list": "ParameterValueList",
            "Feature": "Feature",
            "Tool": "Tool",
            "Tool feature": "ToolFeature",
            "Tool feature method": "ToolFeatureMethod",
        }[new_type]
        root_mapping = import_mapping_from_dict({"map_type": map_type})
        self.set_root_mapping(table_row, list_row, root_mapping)

    def _set_mapping_data(self, flattened_mappings, index, value, role):
        """Sets data for mapping components.

        Args:
            flattened_mappings (FlattenedMappings): component's mapping list
            index (QModelIndex): index
            value (Any): value to set
            role (int): Qt's data role

        Returns:
            bool: True if data was set successfully, False otherwise
        """
        if role != Qt.EditRole:
            return False
        column = index.column()
        if column < FlattenedColumn.POSITION_TYPE:
            return False
        row = index.row()
        current_type = flattened_mappings.display_position_type(row)
        current_position = flattened_mappings.display_position(row)
        if column == FlattenedColumn.POSITION_TYPE:
            if value == current_type:
                return False
            parent_index = index.parent()
            mapping_list_row = parent_index.row()
            table_row = parent_index.parent().row()
            self._undo_stack.push(
                SetMappingPositionType(table_row, mapping_list_row, row, self, value, current_type, current_position)
            )
            return True
        if column == FlattenedColumn.POSITION:
            if current_type != "None":
                if value == current_position:
                    return False
                parent_index = index.parent()
                mapping_list_row = parent_index.row()
                table_row = parent_index.parent().row()
                self._undo_stack.push(
                    SetMappingPosition(
                        table_row, mapping_list_row, row, self, current_type, value, current_type, current_position
                    )
                )
                return True
            # If type is "None", set it to something reasonable to try and help users
            try:
                value = int(value)
            except ValueError:
                pass
            if isinstance(value, int):
                return self.change_component_mapping(flattened_mappings, index, "Column", value)
            elif isinstance(value, str):
                return self.change_component_mapping(flattened_mappings, index, "Constant", value)
        if column == FlattenedColumn.REGEXP:
            parent_index = index.parent()
            mapping_list_row = parent_index.row()
            table_row = parent_index.parent().row()
            current_re = flattened_mappings.component_at(row).filter_re
            self._undo_stack.push(SetFilterRe(table_row, mapping_list_row, row, self, value, current_re))
            return True
        return False

    def get_set_data_delayed(self, index):
        """Returns a function that ParameterValueEditor can call to set data for the given index at any later time,
        even if the model changes.

        Args:
            index (QModelIndex): model index

        Returns:
            Callable: callback for parameter value editor
        """
        return lambda value_type_tup, index=index: self.setData(index, join_value_and_type(*value_type_tup))

    @staticmethod
    def index_name(index):
        """Returns identifier for parameter value editor's title.

        Returns:
            str: 'index' name
        """
        return index.siblingAtColumn(0).data()

    def set_relationship_dimension_count(self, table_row, list_row, dimension_count):
        """Sets relationship dimension count.

        Args:
            table_row (int): source table row index
            list_row (int): mapping list row index
            dimension_count (int): dimension count
        """
        table_index = self.index(table_row, 0)
        list_index = self.index(list_row, 0, table_index)
        mapping_list = self._mappings[table_row].mapping_list[list_row]
        flattened_mappings = mapping_list.flattened_mappings
        self.beginRemoveRows(list_index, 0, len(flattened_mappings.display_names) - 1)
        mapping_list.flattened_mappings = None
        self.endRemoveRows()
        flattened_mappings.set_dimension_count(dimension_count)
        self.beginInsertRows(list_index, 0, len(flattened_mappings.display_names) - 1)
        mapping_list.flattened_mappings = flattened_mappings
        self.endInsertRows()
        self.dataChanged.emit(list_index, list_index, [Role.FLATTENED_MAPPINGS])

    def set_map_dimension_count(self, table_row, list_row, dimension_count):
        """
        Sets map dimension_count.

        Args:
            table_row (int): source table row index
            list_row (int): mapping list row index
            dimension_count (int): new map dimension_count
        """
        table_index = self.index(table_row, 0)
        list_index = self.index(list_row, 0, table_index)
        mapping_list = self._mappings[table_row].mapping_list[list_row]
        flattened_mappings = mapping_list.flattened_mappings
        self.beginRemoveRows(list_index, 0, len(flattened_mappings.display_names) - 1)
        mapping_list.flattened_mappings = None
        self.endRemoveRows()
        flattened_mappings.set_map_dimension_count(dimension_count)
        self.beginInsertRows(list_index, 0, len(flattened_mappings.display_names) - 1)
        mapping_list.flattened_mappings = flattened_mappings
        self.endInsertRows()
        self.dataChanged.emit(list_index, list_index, [Role.FLATTENED_MAPPINGS])

    def set_map_compress(self, table_row, list_row, compress):
        """
        Sets the compress flag for Map type parameters.

        Args:
            table_row (int): source table row index
            list_row (int): mapping list row index
            compress (bool): flag value
        """
        flattened_mappings = self._mappings[table_row].mapping_list[list_row].flattened_mappings
        flattened_mappings.set_map_compress(compress)
        table_index = self.index(table_row, 0)
        list_index = self.index(list_row, 0, table_index)
        self.dataChanged.emit(list_index, list_index, [Role.FLATTENED_MAPPINGS])

    def set_parameter_type(self, table_row, list_row, new_type):
        """Changes parameter type.

        Args:
            table_row (int): source table row index
            list_row (int): mapping list row index
            new_type (str): parameter type
        """
        if new_type == "None":
            parameter_root = None
        elif new_type == "Value":
            parameter_root = parameter_mapping_from_dict({"map_type": "ParameterValue"})
        elif new_type == "Definition":
            parameter_root = parameter_mapping_from_dict({"map_type": "ParameterDefinition"})
        else:
            raise RuntimeError(f"Unknown parameter type '{new_type}'")
        table_index = self.index(table_row, 0)
        list_index = self.index(list_row, 0, table_index)
        mapping_list = self._mappings[table_row].mapping_list[list_row]
        flattened_mappings = mapping_list.flattened_mappings
        self.beginRemoveRows(list_index, 0, len(flattened_mappings.display_names) - 1)
        mapping_list.flattened_mappings = None
        self.endRemoveRows()
        flattened_mappings.set_parameter_components(parameter_root)
        self.beginInsertRows(list_index, 0, len(flattened_mappings.display_names) - 1)
        mapping_list.flattened_mappings = flattened_mappings
        self.endInsertRows()
        self.dataChanged.emit(list_index, list_index, [Role.FLATTENED_MAPPINGS])

    def set_value_type(self, table_row, list_row, new_type):
        """Changes parameter value's type.

        Args:
            table_row (int): source table row index
            list_row (int): mapping list row index
            new_type (str): value type
        """
        table_index = self.index(table_row, 0)
        list_index = self.index(list_row, 0, table_index)
        mapping_list = self._mappings[table_row].mapping_list[list_row]
        flattened_mappings = mapping_list.flattened_mappings
        if new_type == "None":
            value_mappings = None
        else:
            value_type = VALUE_TYPES[new_type]
            parameter_type = flattened_mappings.display_parameter_type()
            if parameter_type == "Definition":
                value_mappings = parameter_default_value_mapping_from_dict({"value_type": value_type})
            elif parameter_type == "Value":
                value_mappings = parameter_value_mapping_from_dict({"value_type": value_type})
            else:
                raise RuntimeError(f"Unknown value type '{value_type}'")
        self.beginRemoveRows(list_index, 0, len(flattened_mappings.display_names) - 1)
        mapping_list.flattened_mappings = None
        self.endRemoveRows()
        flattened_mappings.set_value_components(value_mappings)
        self.beginInsertRows(list_index, 0, len(flattened_mappings.display_names) - 1)
        mapping_list.flattened_mappings = flattened_mappings
        self.endInsertRows()
        self.dataChanged.emit(list_index, list_index, [Role.FLATTENED_MAPPINGS])

    def change_component_mapping(self, flattened_mappings, index, new_type, new_ref):
        """
        Pushes :class:`SetComponentMappingType` to the undo stack.

        Args:
            flattened_mappings (FlattenedMappings): flattened mappings
            index (QModelIndex): flattened mappings index
            new_type (str): name of the new type
            new_ref (str or int): component mapping reference

        Returns:
            bool: True if the operation was successful, False otherwise
        """
        row = index.row()
        previous_type = flattened_mappings.display_position_type(row)
        previous_ref = flattened_mappings.display_position(row)
        parent_index = index.parent()
        mapping_list_row = parent_index.row()
        table_row = parent_index.parent().row()
        commands = list()
        if new_type != previous_type:
            commands.append(
                SetMappingPositionType(table_row, mapping_list_row, row, self, new_type, previous_type, previous_ref)
            )
        if new_ref != previous_ref:
            commands.append(
                SetMappingPosition(
                    table_row, mapping_list_row, row, self, new_type, new_ref, previous_type, previous_ref
                )
            )
        if not commands:
            return False
        self._undo_stack.beginMacro("mapping type and reference change")
        for command in commands:
            self._undo_stack.push(command)
        self._undo_stack.endMacro()
        return True

    def set_mapping_position_type(self, table_row, list_row, row, position_type):
        """Modifies mapping component's position type.

        Args:
            table_row (int): source table row index
            list_row (int): mapping list row index
            row (int): mapping component row index
            position_type (str): new type
        """
        flattened_mappings = self._mappings[table_row].mapping_list[list_row].flattened_mappings
        flattened_mappings.set_display_position_type(row, position_type)
        table_index = self.index(table_row, 0)
        list_index = self.index(list_row, 0, table_index)
        top_left = self.index(row, 1, list_index)
        bottom_right = self.index(self.rowCount(list_index) - 1, 2, list_index)
        self.dataChanged.emit(top_left, bottom_right, [Qt.BackgroundRole, Qt.DisplayRole, Qt.ToolTipRole])

    def set_mapping_position(self, table_row, list_row, row, position_type, position):
        """Modifies mapping component's position

        Args:
            table_row (int): source table row index
            list_row (int): mapping list row index
            row (int): mapping component row index
            position_type (str): new type
            position (str or int): new position
        """
        flattened_mappings = self._mappings[table_row].mapping_list[list_row].flattened_mappings
        flattened_mappings.set_display_position(row, position_type, position)
        table_index = self.index(table_row, 0)
        list_index = self.index(list_row, 0, table_index)
        top_left = self.index(row, 1, list_index)
        bottom_right = self.index(self.rowCount(list_index) - 1, 2, list_index)
        self.dataChanged.emit(top_left, bottom_right, [Qt.BackgroundRole, Qt.DisplayRole, Qt.ToolTipRole])
        self._recommend_source_table_column_or_row_types(flattened_mappings, row)

    def set_import_objects(self, table_row, list_row, import_objects):
        """Sets the import object flag.

        Args:
            table_row (int): source table row index
            list_row (int): mapping list row index
            import_objects (bool): flag value
        """
        flattened_mappings = self._mappings[table_row].mapping_list[list_row].flattened_mappings
        flattened_mappings.set_import_objects(import_objects)
        table_index = self.index(table_row, 0)
        list_index = self.index(list_row, 0, table_index)
        self.dataChanged.emit(list_index, list_index, [Role.FLATTENED_MAPPINGS])

    def set_use_before_alternative(self, table_row, list_row, use_before_alternative):
        """Sets the use before alternative flag.

        Args:
            table_row (int): source table row index
            list_row (int): mapping list row index
            use_before_alternative (bool): flag value
        """
        table_index = self.index(table_row, 0)
        list_index = self.index(list_row, 0, table_index)
        mapping_list = self._mappings[table_row].mapping_list[list_row]
        flattened_mappings = mapping_list.flattened_mappings
        if use_before_alternative:
            before_alternative_mapping = ScenarioBeforeAlternativeMapping(Position.hidden)
            self.beginInsertRows(list_index, 2, 2)
            flattened_mappings.append_tail_component(before_alternative_mapping)
            self.endInsertRows()
        else:
            self.beginRemoveRows(list_index, 2, 2)
            flattened_mappings.cut_tail_component()
            self.endRemoveRows()
        self.dataChanged.emit(list_index, list_index, [Role.FLATTENED_MAPPINGS])

    def set_read_start_row(self, table_row, list_row, start_row):
        """Sets mappings' read start row.

        Args:
            table_row (int): source table row index
            list_row (int): mapping list row index
            start_row (int): read start row
        """
        flattened_mappings = self._mappings[table_row].mapping_list[list_row].flattened_mappings
        flattened_mappings.set_read_start_row(start_row)
        table_index = self.index(table_row, 0)
        list_index = self.index(list_row, 0, table_index)
        self.dataChanged.emit(list_index, list_index, [Role.FLATTENED_MAPPINGS])

    def set_skip_columns(self, table_row, list_row, skip_columns):
        """Sets mappings' read start row.

        Args:
            table_row (int): source table row index
            list_row (int): mapping list row index
            skip_columns (list): skipped column indexes or names
        """
        if skip_columns is None:
            skip_columns = []
        flattened_mappings = self._mappings[table_row].mapping_list[list_row].flattened_mappings
        flattened_mappings.set_skip_columns(sorted(list(set(skip_columns))))
        table_index = self.index(table_row, 0)
        list_index = self.index(list_row, 0, table_index)
        self.dataChanged.emit(list_index, list_index, [Role.FLATTENED_MAPPINGS])

    def set_time_series_repeat_flag(self, table_row, list_row, repeat):
        """Sets the repeat time series flag.

        Args:
            table_row (int): source table row index
            list_row (int): mapping list row index
            repeat (bool): flag value
        """
        flattened_mappings = self._mappings[table_row].mapping_list[list_row].flattened_mappings
        if not flattened_mappings.is_time_series_value():
            return
        flattened_mappings.value_mapping.options["repeat"] = repeat
        table_index = self.index(table_row, 0)
        list_index = self.index(list_row, 0, table_index)
        self.dataChanged.emit(list_index, list_index, [Role.FLATTENED_MAPPINGS])

    def set_filter_re(self, table_row, list_row, row, filter_re):
        """Sets filter regular expression.

        Args:
            table_row (int): source table row index
            list_row (int): mapping list row index
            row (int): mapping component row index
            filter_re (str): filter regular expression
        """
        flattened_mappings = self._mappings[table_row].mapping_list[list_row].flattened_mappings
        flattened_mappings.component_at(row).filter_re = filter_re
        table_index = self.index(table_row, 0)
        list_index = self.index(list_row, 0, table_index)
        index = self.index(row, FlattenedColumn.REGEXP, list_index)
        self.dataChanged.emit(index, index, [Qt.DisplayRole])

    @staticmethod
    def polish_mapping(list_index, header):
        """Polishes flattened mappings.

        Args:
            list_index (QModelIndex): index to mappings list
            header (Iterable): source table header
        """
        mapping_list_item = list_index.internalPointer()
        table_name = mapping_list_item.source_table_item.name
        mapping_list_item.flattened_mappings.root_mapping.polish(table_name, header, for_preview=True)
        # We don't emit dataChanged here as polishing is just beautification and
        # it would mess up the undo system which relies on real changes.

    def _recommend_source_table_column_or_row_types(self, flattened_mappings, row):
        """Checks if source table conversion specifications need an update.

        Args:
            flattened_mappings (FlattenedMappings): flattened mappings
            row (int): display row index
        """
        self._recommend_float_type_for_values(flattened_mappings)
        display_name = flattened_mappings.display_names[row]
        if display_name.endswith("indexes") and flattened_mappings.is_time_series_value():
            self._recommend_date_time_type(flattened_mappings.component_at(row))
        elif not display_name.endswith("values"):
            self._recommend_string_type(flattened_mappings.component_at(row))

    def _recommend_float_type_for_values(self, flattened_mappings):
        """

        Args:
            flattened_mappings (FlattenedMappings): flattened mappings
        """
        if not flattened_mappings.has_value_component():
            return
        visual_row = next(
            (k for k, name in enumerate(flattened_mappings.display_names) if name.endswith("values")), None
        )
        if visual_row is None:
            return
        component = flattened_mappings.component_at(visual_row)
        if not flattened_mappings.root_mapping.is_pivoted():
            self._recommend_float_type(component)
            return
        excluded_columns = flattened_mappings.root_mapping.non_pivoted_columns()
        self.multi_column_type_recommended.emit(excluded_columns, FloatConvertSpec())

    def _recommend_string_type(self, component):
        """Signals a need to change source table conversion specification to StringConvertSpec.

        Args:
            component (ImportMapping): mapping component
        """
        self._recommend_type(component, StringConvertSpec())

    def _recommend_float_type(self, component):
        """Signals a need to change source table conversion specification to FloatConvertSpec.

        Args:
            component (ImportMapping): mapping component
        """
        self._recommend_type(component, FloatConvertSpec())

    def _recommend_date_time_type(self, component):
        """Signals a need to change source table conversion specification to DateTimeConvertSpec.

        Args:
            component (ImportMapping): mapping component
        """
        self._recommend_type(component, DateTimeConvertSpec())

    def _recommend_type(self, component, convert_spec):
        """Signals a need to change source table conversion specifications.

        Args:
            component (ImportMapping): mapping component
            convert_spec (ConvertSpec): row/column conversion specification
        """
        if not isinstance(component.position, int):
            return
        if component.position >= 0:
            self.row_or_column_type_recommended.emit(component.position, convert_spec, Qt.Horizontal)
        else:
            self.row_or_column_type_recommended.emit(-(component.position + 1), convert_spec, Qt.Vertical)

    def set_tables_editable(self, editable):
        """Enables or disables table name editing.

        Args:
            editable (bool): True to enable table name editing, False to disable
        """
        for item in self._mappings[1:]:
            if item.name == UNNAMED_TABLE_NAME:
                continue
            item.editable = editable

    def add_empty_row(self):
        """Adds the special 'unnamed table' row at the end of table list."""
        last_row = len(self._mappings) - 1
        empty_item = SourceTableItem(UNNAMED_TABLE_NAME, checked=False, checkable=False, editable=True, real=False)
        self.beginInsertRows(QModelIndex(), last_row, last_row)
        self._mappings.append(empty_item)
        self.endInsertRows()

    def insertRow(self, row, parent=QModelIndex()):
        parent_item = parent.internalPointer()
        if isinstance(parent_item, SourceTableItem) and parent_item.real:
            list_item = MappingListItem(self._unique_mapping_name(parent_item))
            list_item.set_flattened_mappings(FlattenedMappings(self._create_default_mapping()))
            self.beginInsertRows(parent, row, row)
            parent_item.insert_to_mapping_list(row, list_item)
            self.endInsertRows()
            return True
        return False

    def removeRow(self, row, parent=QModelIndex()):
        parent_item = parent.internalPointer()
        if parent_item is None:
            self.beginRemoveRows(parent, row, row)
            self._mappings.pop(row)
            self.endRemoveRows()
            return True
        if isinstance(parent_item, SourceTableItem):
            self.beginRemoveRows(parent, row, row)
            parent_item.mapping_list.pop(row)
            self.endRemoveRows()
            return True
        return False

    def removeRows(self, row, count, parent=QModelIndex()):
        parent_item = parent.internalPointer()
        if parent_item is None:
            self.beginRemoveRows(parent, row, row + count - 1)
            self._mappings = self._mappings[:row] + self._mappings[row + count :]
            self.endRemoveRows()
            return True
        if isinstance(parent_item, SourceTableItem):
            self.beginRemoveRows(parent, row, row + count - 1)
            parent_item.mapping_list = parent_item.mapping_list[:row] + parent_item.mapping_list[row + count :]
            self.endRemoveRows()
            return True
        return False

    @staticmethod
    def _create_default_mapping():
        """Creates a default object class mapping.

        Returns:
            ImportMapping: new mapping root
        """
        return import_mapping_from_dict({"map_type": "ObjectClass"})

    def has_mapping_name(self, table_row, name):
        """Checks if a name exists in mapping list.

        Args:
            table_row (int): source table row index
            name (str): name to check

        Returns:
            bool: True if name exists, False otherwise
        """
        return name in {i.name for i in self._mappings[table_row].mapping_list}

    def unique_mapping_name(self, table_row, prefix):
        """Generates a unique mapping name.

        Args:
            table_row (int): source table row index
            prefix (str): name prefix

        Returns:
            str: name
        """
        return self._unique_mapping_name(self._mappings[table_row], prefix)

    @staticmethod
    def _unique_mapping_name(table_item, prefix="Mapping"):
        """Generates a unique mapping name.

        Args:
            table_item (SourceTableItem): source table item
            prefix (str): name prefix

        Returns:
            str: name
        """
        return unique_name(prefix, [i.name for i in table_item.mapping_list])



######################################################################################################################
# Copyright (C) 2017-2021 Spine project consortium
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""
Exporter's execute kernel (do_work), as target for a multiprocess.Process

:authors: A. Soininen (VTT)
:date:    14.12.2020
"""
from pathlib import Path
from spinedb_api.spine_io.exporters.writer import write, WriterException
from spinedb_api.spine_io.exporters.csv_writer import CsvWriter
from spinedb_api.spine_io.exporters.excel_writer import ExcelWriter
from spinedb_api.spine_io.exporters.gdx_writer import GdxWriter
from spinedb_api.spine_io.exporters.sql_writer import SqlWriter
from spinedb_api import DatabaseMapping, SpineDBAPIError
from spine_engine.utils.helpers import write_filter_id_file
from spine_items.utils import subdirectory_for_fork
from .specification import Specification, OutputFormat


def do_work(
    specification, output_time_stamps, cancel_on_error, gams_path, out_dir, databases, filter_id, filter_id_hash, logger
):
    """
    Exports databases using given specification as export mapping.

    Args:
        specification (dict): export specification dictionary
        output_time_stamps (bool): if True, puts output files into time stamped subdirectories
        cancel_on_error (bool): if True, bails out on non-fatal errors
        gams_path (str): path to GAMS installation
        out_dir (str): base output directory
        databases (dict): databases to export
        filter_id (str): filter id
        filter_id_hash (str): hashed filter id
        logger (LoggerInterface): a logger

    Returns:
        tuple: boolean success flag, dictionary of output files
    """
    specification = Specification.from_dict(specification)
    successes = list()
    written_files = dict()
    for url, output_file_name in databases.items():
        out_path = subdirectory_for_fork(output_file_name, out_dir, output_time_stamps, filter_id_hash)
        try:
            database_map = DatabaseMapping(url)
        except SpineDBAPIError as error:
            logger.msg_error.emit(f"Failed to export <b>{url}</b>: {error}")
            if cancel_on_error:
                return False, written_files
            successes.append(False)
            continue
        try:
            try:
                file = Path(out_path)
                file.parent.mkdir(parents=True, exist_ok=True)
                if file.exists():
                    file.unlink()
                writer = make_writer(specification.output_format, out_path, gams_path)
                specifications = specification.enabled_specifications().values()
                mappings = (m.root for m in specifications)
                header_always = (m.always_export_header for m in specifications)
                group_fns = (m.group_fn for m in specifications)
                write(database_map, writer, *mappings, empty_data_header=header_always, group_fns=group_fns)
            except (PermissionError, WriterException) as e:
                logger.msg_error.emit(str(e))
                if cancel_on_error:
                    return False, written_files
                successes.append(False)
            else:
                if isinstance(writer, CsvWriter):
                    files = writer.output_files()
                else:
                    files = {out_path}
                written_files[output_file_name] = files
                if len(files) > 1:
                    anchors = list()
                    for path in (Path(f) for f in files):
                        anchors.append(
                            f"<a style='color:#BB99FF;' title='{path}' href='file:///{path}'>{path.name}</a>"
                        )
                    logger.msg_success.emit(f"Wrote multiple files:<br>{'<br>'.join(anchors)}")
                else:
                    file_anchor = f"<a style='color:#BB99FF;' title='{file}' href='file:///{file}'>{file.name}</a>"
                    logger.msg_success.emit(f"File {file_anchor} written.")
                if filter_id:
                    write_filter_id_file(filter_id, Path(out_path).parent)
                successes.append(True)
        finally:
            database_map.connection.close()
    return all(successes), written_files


def make_writer(output_format, out_path, gams_path):
    """
    Constructs a writer.

    Args:
        output_format (OutputFormat): output format
        out_path (str): path to output file
        gams_path (str): path to GAMS installation

    Returns:
        Writer: a writer
    """
    if output_format == OutputFormat.CSV:
        path = Path(out_path)
        return CsvWriter(path.parent, path.name)
    elif output_format == OutputFormat.EXCEL:
        return ExcelWriter(out_path)
    elif output_format == OutputFormat.SQL:
        return SqlWriter(out_path)
    return GdxWriter(out_path, gams_path)



######################################################################################################################
# Copyright (C) 2017-2021 Spine project consortium
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""
Contains a simple filter widget for skip column filtering in mapping options.

:author: A. Soininen (VTT)
:date:   20.10.2021
"""
from spinetoolbox.mvcmodels.filter_checkbox_list_model import DataToValueFilterCheckboxListModel
from spinetoolbox.widgets.custom_qwidgets import FilterWidgetBase


class SimpleFilterWidget(FilterWidgetBase):
    def __init__(self, parent, show_empty=True):
        """
        Args:
            parent (QWidget): parent widget
        """
        super().__init__(parent)
        self._filter_model = DataToValueFilterCheckboxListModel(self, str, show_empty=show_empty)
        self._filter_model.set_list(self._filter_state)
        self._ui_list.setModel(self._filter_model)
        self.connect_signals()



######################################################################################################################
# Copyright (C) 2017-2021 Spine project consortium
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""
Contains base class for project items.

:author: A. Soininen (VTT)
:date:   15.12.2020
"""

from PySide2.QtCore import Qt, Slot
from spinetoolbox.project_item.project_item import ProjectItem
from .commands import UpdateCancelOnErrorCommand, UpdateOutputTimeStampsFlag
from .models import FullUrlListModel


class ExporterBase(ProjectItem):
    """A base class for exporter project items."""

    def __init__(self, name, description, x, y, toolbox, project, output_time_stamps=False, cancel_on_error=True):
        """
        Args:
            name (str): item name
            description (str): item description
            x (float): initial X coordinate of item icon
            y (float): initial Y coordinate of item icon
            toolbox (ToolboxUI): a ToolboxUI instance
            project (SpineToolboxProject): the project this item belongs to
            output_time_stamps (bool): True if time stamps should be appended to output directory names,
                False otherwise
            cancel_on_error (bool): True if execution should fail on all export errors,
                False to ignore certain error cases; optional to provide backwards compatibility
        """
        super().__init__(name, description, x, y, project)
        self._toolbox = toolbox
        self._append_output_time_stamps = output_time_stamps
        self._cancel_on_error = cancel_on_error
        self._output_filenames = dict()
        self._export_list_items = dict()
        self._full_url_model = FullUrlListModel()
        self._exported_files = None

    @staticmethod
    def item_type():
        """See base class."""
        raise NotImplementedError()

    @staticmethod
    def item_category():
        """See base class."""
        raise NotImplementedError()

    @property
    def executable_class(self):
        raise NotImplementedError()

    def handle_execution_successful(self, execution_direction, engine_state):
        """See base class."""
        if execution_direction != "FORWARD":
            return
        self._resources_to_successors_changed()

    def full_url_model(self):
        """
        Returns the full URL model held by the exporter.

        Returns:
            FullUrlListModel: full URL model
        """
        return self._full_url_model

    def make_signal_handler_dict(self):
        """Returns a dictionary of all shared signals and their handlers."""
        s = super().make_signal_handler_dict()
        s[self._properties_ui.output_time_stamps_check_box.stateChanged] = self._change_output_time_stamps_flag
        s[self._properties_ui.cancel_on_error_check_box.stateChanged] = self._cancel_on_error_option_changed
        return s

    def restore_selections(self):
        """Restores selections and connects signals."""
        self._update_properties_tab()

    @Slot(int)
    def _change_output_time_stamps_flag(self, checkbox_state):
        """
        Pushes a command that changes the output time stamps flag value.

        Args:
            checkbox_state (int): setting's checkbox state on properties tab
        """
        flag = checkbox_state == Qt.Checked
        if flag == self._append_output_time_stamps:
            return
        self._toolbox.undo_stack.push(UpdateOutputTimeStampsFlag(self, flag))

    def set_output_time_stamps_flag(self, flag):
        """
        Sets the output time stamps flag.

        Args:
            flag (bool): flag value
        """
        self._append_output_time_stamps = flag

    @Slot(int)
    def _cancel_on_error_option_changed(self, checkbox_state):
        """Handles changes to the Cancel export on error option."""
        cancel = checkbox_state == Qt.Checked
        if self._cancel_on_error == cancel:
            return
        self._toolbox.undo_stack.push(UpdateCancelOnErrorCommand(self, cancel))

    def set_cancel_on_error(self, cancel):
        """Sets the Cancel export on error option."""
        self._cancel_on_error = cancel
        if not self._active:
            return
        # This does not trigger the stateChanged signal.
        self._properties_ui.cancel_on_error_check_box.setCheckState(Qt.Checked if cancel else Qt.Unchecked)

    def item_dict(self):
        """Returns a dictionary corresponding to this item's configuration."""
        d = super().item_dict()
        d["output_time_stamps"] = self._append_output_time_stamps
        d["cancel_on_error"] = self._cancel_on_error
        return d

    @staticmethod
    def from_dict(name, item_dict, toolbox, project):
        """See base class"""
        raise NotImplementedError()

    def tear_down(self):
        super().tear_down()
        self._full_url_model.deleteLater()



######################################################################################################################
# Copyright (C) 2017-2021 Spine project consortium
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""
A small widget to set up an output in Exporter properties tab.

:author: A. Soininen (VTT)
:date:   10.9.2019
"""

from PySide2.QtCore import Signal, Slot
from PySide2.QtWidgets import QWidget


class ExportListItem(QWidget):
    """A widget with few controls to select the export output label."""

    out_label_changed = Signal(str, str)
    """Emitted when the output label field is changed."""

    def __init__(self, in_label, out_label, parent=None):
        """
        Args:
            in_label (str): input resource label
            out_label (str): output resource label
            parent (QWidget): a parent widget
        """
        from ..ui.export_list_item import Ui_Form  # pylint: disable=import-outside-toplevel

        super().__init__(parent)
        self._ui = Ui_Form()
        self._in_label = in_label
        self._out_label = out_label
        self._ui.setupUi(self)
        self._ui.input_label_field.setText(f"<b>{in_label}</b>")
        self._ui.out_label_edit.setText(out_label)
        self._ui.out_label_edit.editingFinished.connect(self._emit_out_label_changed)

    @property
    def out_label_edit(self):
        """Output label QLineEdit"""
        return self._ui.out_label_edit

    @property
    def in_label_field(self):
        """Input label QLineEdit"""
        return self._ui.input_label_field

    @Slot()
    def _emit_out_label_changed(self):
        """Emits out_label_changed signal."""
        out_label = self._ui.out_label_edit.text()
        if self._out_label == out_label:
            return
        self._out_label = out_label
        self.out_label_changed.emit(out_label, self._in_label)



######################################################################################################################
# Copyright (C) 2017-2021 Spine project consortium
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""
GdxExporter's execute kernel (do_work), as target for a multiprocess.Process

:authors: M. Marin (KTH)
:date:    6.11.2020
"""
import os
from spinedb_api.spine_io.exporters import gdx
from spinedb_api import clear_filter_configs, DatabaseMapping, SpineDBAPIError
from spinedb_api.filters.tools import ensure_filtering
from spine_items.utils import subdirectory_for_fork


def do_work(
    settings_pack, output_time_stamps, cancel_on_error, data_dir, gams_system_directory, databases, forks, logger
):
    successes = list()
    written_files = dict()
    for url, output_file_name in databases.items():
        out_path = subdirectory_for_fork(
            output_file_name, data_dir, output_time_stamps, forks[clear_filter_configs(url)]
        )
        try:
            database_map = DatabaseMapping(ensure_filtering(url, fallback_alternative="Base"))
        except SpineDBAPIError as error:
            logger.msg_error.emit(f"Failed to export <b>{url}</b> to .gdx: {error}")
            successes.append(False)
            continue
        try:
            export_logger = logger if not cancel_on_error else None
            try:
                os.mkdir(os.path.dirname(out_path))
            except FileExistsError:
                pass
            try:
                gdx.to_gdx_file(
                    database_map,
                    out_path,
                    settings_pack.settings,
                    settings_pack.indexing_settings,
                    settings_pack.merging_settings,
                    settings_pack.none_fallback,
                    settings_pack.none_export,
                    gams_system_directory,
                    export_logger,
                )
            except gdx.GdxExportException as error:
                logger.msg_error.emit(f"Failed to export <b>{url}</b> to .gdx: {error}")
                successes.append(False)
                continue
            written_files[output_file_name] = [out_path]
            logger.msg_success.emit(f"File <b>{out_path}</b> written")
            successes.append(True)
        finally:
            database_map.connection.close()
    return (all(successes), written_files)



######################################################################################################################
# Copyright (C) 2017-2021 Spine project consortium
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""
Contains the GdxExporter project item.

:author: A. Soininen (VTT)
:date:   5.9.2019
"""

from copy import deepcopy
from itertools import zip_longest
import os.path
from PySide2.QtCore import Slot, Qt
from spinetoolbox.project_item.project_item import ProjectItem
from spine_engine.utils.serialization import deserialize_path, serialize_url
from spinedb_api import clear_filter_configs
from spinedb_api.spine_io.exporters import gdx
from ..commands import UpdateOutFileName
from ..models import DatabaseListModel
from ..utils import Database, exported_files_as_resources
from ..item_base import ExporterBase
from .commands import UpdateSettings
from .executable_item import ExecutableItem
from .item_info import ItemInfo
from .settings_pack import SettingsPack
from .settings_state import SettingsState
from .widgets.gdx_export_settings import GdxExportSettings
from .widgets.export_list_item import ExportListItem
from .worker import Worker
from .exporter_notifications import ExporterNotifications


class GdxExporter(ExporterBase):
    """
    This project item handles all functionality regarding exporting a database to a .gdx file.
    """

    def __init__(
        self,
        name,
        description,
        x,
        y,
        toolbox,
        project,
        settings_pack=None,
        databases=None,
        output_time_stamps=False,
        cancel_on_error=True,
    ):
        """
        Args:
            name (str): item name
            description (str): item description
            x (float): initial X coordinate of item icon
            y (float): initial Y coordinate of item icon
            toolbox (ToolboxUI): a ToolboxUI instance
            project (SpineToolboxProject): the project this item belongs to
            settings_pack (SettingsPack, optional): export settings
            databases (list, optional): a list of :class:`Database` instances
            output_time_stamps (bool): True if time stamps should be appended to output directory names,
                False otherwise
            cancel_on_error (bool): True if execution should fail on all export errors,
                False to ignore certain error cases; optional to provide backwards compatibility
        """
        super().__init__(name, description, x, y, toolbox, project, output_time_stamps, cancel_on_error)
        self._notifications = ExporterNotifications()
        self._settings_pack = settings_pack if settings_pack is not None else SettingsPack()
        if databases is None:
            databases = list()
        self._database_model = DatabaseListModel(databases)
        self._inactive_databases = dict()
        self._settings_window = None
        self._worker = None

    @staticmethod
    def item_type():
        """See base class."""
        return ItemInfo.item_type()

    @staticmethod
    def item_category():
        """See base class."""
        return ItemInfo.item_category()

    @property
    def executable_class(self):
        return ExecutableItem

    def settings_pack(self):
        """
        Returns export settings.

        Returns:
            SettingsPack: settings
        """
        return self._settings_pack

    def make_signal_handler_dict(self):
        """Returns a dictionary of all shared signals and their handlers."""
        s = super().make_signal_handler_dict()
        s[self._properties_ui.settings_button.clicked] = self.show_settings
        return s

    def _check_notifications(self):
        """See base class."""
        super()._check_notifications()
        self._check_missing_file_names()
        self._check_duplicate_file_names()
        self._check_missing_specification()
        self._report_notifications()
        self.add_notification(
            "GdxExporter is deprecated and will be removed in a future release. "
            "Please replace this item by general-purpose Exporter."
        )

    def _report_notifications(self):
        """Updates the exclamation icon and notifications labels."""
        if self._icon is None:
            return
        self.clear_notifications()
        if self._notifications.duplicate_output_file_name:
            self.add_notification("Duplicate output file names.")
        if self._notifications.missing_output_file_name:
            self.add_notification("Output file name(s) missing.")
        if self._notifications.missing_specification:
            self.add_notification("Export specification missing.")

    def _check_missing_specification(self):
        self._notifications.missing_specification = (
            self._database_model.rowCount() != 0 and self._settings_pack.settings is None
        )

    def _check_missing_file_names(self):
        """Checks the status of output file names."""
        self._notifications.missing_output_file_name = not all(
            bool(db.output_file_name) for db in self._database_model.items()
        )

    def _check_duplicate_file_names(self):
        """Checks for duplicate output file names."""
        self._notifications.duplicate_output_file_name = False
        names = set()
        for db in self._database_model.items():
            if db.output_file_name in names:
                self._notifications.duplicate_output_file_name = True
                break
            names.add(db.output_file_name)

    def database(self, url):
        """
        Returns database information for given URL.

        Args:
            url (str): database URL

        Returns:
            Database: database information
        """
        return self._database_model.item(url)

    def _update_properties_tab(self):
        """Updates the database list in the properties tab."""
        database_list_storage = self._properties_ui.databases_list_layout
        while not database_list_storage.isEmpty():
            widget_to_remove = database_list_storage.takeAt(0)
            widget_to_remove.widget().deleteLater()
        self._export_list_items.clear()
        for db in self._database_model.items():
            item = self._export_list_items[db.url] = ExportListItem(db.url, db.output_file_name)
            database_list_storage.addWidget(item)
            item.file_name_changed.connect(self._update_out_file_name)
        self._properties_ui.output_time_stamps_check_box.setCheckState(
            Qt.Checked if self._append_output_time_stamps else Qt.Unchecked
        )
        self._properties_ui.cancel_on_error_check_box.setCheckState(
            Qt.Checked if self._cancel_on_error else Qt.Unchecked
        )

    def upstream_resources_updated(self, resources):
        """See base class."""
        full_urls = set(r.url for r in resources if r.type_ == "database")
        database_urls = set(clear_filter_configs(url) for url in full_urls)
        old_urls = self._database_model.urls()
        if database_urls != old_urls:
            common = old_urls & database_urls
            old_urls_by_base = dict()
            for url in old_urls:
                if url not in common:
                    old_urls_by_base.setdefault(clear_filter_configs(url), list()).append(url)
            old_urls_by_base = {base: sorted(url_list) for base, url_list in old_urls_by_base.items()}
            new_urls_by_base = dict()
            for url in database_urls:
                if url not in common:
                    new_urls_by_base.setdefault(clear_filter_configs(url), list()).append(url)
            new_urls_by_base = {base: sorted(url_list) for base, url_list in new_urls_by_base.items()}
            useless_oldies = set()
            homeless_new_urls = set()
            for new_base, newbies in new_urls_by_base.items():
                oldies = old_urls_by_base.get(new_base)
                if oldies is None:
                    homeless_new_urls |= set(newbies)
                    continue
                for oldie, newbie in zip_longest(oldies, newbies):
                    if oldie is None:
                        homeless_new_urls.add(newbie)
                        continue
                    if newbie is None:
                        useless_oldies.add(oldie)
                        continue
                    self._database_model.update_url(oldie, newbie)
            useless_oldies |= old_urls - database_urls
            for url in useless_oldies:
                self._inactive_databases[url] = self._database_model.remove(url)
            for url in homeless_new_urls:
                db = self._inactive_databases.pop(url, None)
                if db is None:
                    db = Database()
                    db.url = url
                self._database_model.add(db)
        self._full_url_model.set_urls(full_urls)
        if self._active:
            self._update_properties_tab()
        self._check_notifications()

    def replace_resources_from_upstream(self, old, new):
        """See base class."""
        for old_resource, new_resource in zip(old, new):
            if not old_resource.type_ == "database":
                continue
            self._database_model.update_url(
                clear_filter_configs(old_resource.url), clear_filter_configs(new_resource.url)
            )
            self._full_url_model.update_url(old_resource.url, new_resource.url)

    def set_out_file_name(self, file_name, database_path):
        """Updates the output file name for given database"""
        if self._active:
            export_list_item = self._export_list_items[database_path]
            export_list_item.out_file_name_edit.setText(file_name)
        database = self._database_model.item(database_path)
        old_file_name = database.output_file_name
        old_resources = self.resources_for_direct_successors() if old_file_name else []
        database.output_file_name = file_name
        self._notifications.missing_output_file_name = not file_name
        self._check_duplicate_file_names()
        self._report_notifications()
        if self._exported_files is not None:
            exported_files = self._exported_files.pop(old_file_name, None)
            if exported_files is not None and file_name:
                self._exported_files[file_name] = exported_files
        if not old_file_name or not file_name:
            self._resources_to_successors_changed()
        else:
            new_resources = self.resources_for_direct_successors()
            olds = [resource for resource in old_resources if resource.label == old_file_name]
            news = [resource for resource in new_resources if resource.label == file_name]
            self._resources_to_successors_replaced(olds, news)

    def resources_for_direct_successors(self):
        """See base class."""
        resources, self._exported_files = exported_files_as_resources(
            self.name, self._exported_files, self.data_dir, self._database_model.items()
        )
        return resources

    @Slot(str, str)
    def _update_out_file_name(self, file_name, url):
        """Pushes a new UpdateExporterOutFileNameCommand to the toolbox undo stack."""
        self._toolbox.undo_stack.push(UpdateOutFileName(self, file_name, url))

    def _start_worker(self, database_url, update_settings=False):
        """Starts fetching settings using a worker in another thread."""
        worker = self._worker
        self._worker = None
        if worker is not None:
            worker.thread.quit()
            worker.thread.wait()
            worker.deleteLater()
        worker = Worker(database_url, self._settings_pack.none_fallback, self._logger)
        self._worker = worker
        worker.database_unavailable.connect(self._cancel_worker)
        worker.finished.connect(self._worker_finished)
        worker.errored.connect(self._worker_failed)
        if update_settings:
            worker.set_previous_settings(
                self._settings_window.set_settings,
                self._settings_window.indexing_settings,
                self._settings_window.merging_settings,
            )
        worker.thread.start()

    @Slot(object)
    def _worker_finished(self, result):
        """Gets and updates and export settings pack from a worker."""
        if self._worker is None:
            return
        self._worker.thread.quit()
        self._worker.thread.wait()
        self._worker.deleteLater()
        if self._settings_window is not None:
            self._settings_window.reset_settings(result.set_settings, result.indexing_settings, result.merging_settings)
        else:
            self._settings_pack.settings = result.set_settings
            self._settings_pack.indexing_settings = result.indexing_settings
            self._settings_pack.merging_settings = result.merging_settings
        self._worker = None
        self._toolbox.update_window_modified(False)
        self._check_notifications()

    @Slot(object)
    def _worker_failed(self, exception):
        """Clean up after a worker has failed fetching export settings."""
        if self._worker is None:
            return
        database_url = clear_filter_configs(self._worker.database_url)
        self._worker.thread.quit()
        self._worker.thread.wait()
        self._worker.deleteLater()
        self._worker = None
        if self._settings_window is not None:
            self._settings_window.settings_reading_cancelled()
        self._logger.msg_error.emit(
            f"<b>[{self.name}]</b> Initializing settings for database {database_url} failed: {exception}"
        )
        self._report_notifications()

    @Slot()
    def _cancel_worker(self):
        """Cleans up after worker has given up fetching export settings."""
        if self._worker is None:
            return
        self._worker.thread.quit()
        self._worker.thread.wait()
        self._worker.deleteLater()
        self._worker = None
        if self._settings_window is not None:
            self._settings_window.settings_reading_cancelled()

    @Slot(bool)
    def show_settings(self, _=True):
        """Opens the item's settings window."""
        # Give window its own settings and indexing domains so Cancel doesn't change anything here.
        if self._settings_window is None:
            settings = deepcopy(self._settings_pack.settings)
            indexing_settings = deepcopy(self._settings_pack.indexing_settings)
            merging_settings = deepcopy(self._settings_pack.merging_settings)
            self._settings_window = GdxExportSettings(
                settings,
                indexing_settings,
                merging_settings,
                self._settings_pack.none_fallback,
                self._settings_pack.none_export,
                self._full_url_model,
                self.name,
                self._toolbox,
            )
            self._settings_window.settings_accepted.connect(self._update_settings_from_settings_window)
            self._settings_window.settings_rejected.connect(self._dispose_settings_window)
            self._settings_window.update_requested.connect(self._update_settings_in_window)
        self._settings_window.show()

    @Slot(str)
    def _update_settings_in_window(self, database_url):
        """Updates settings currently in the Gdx Export Settings window."""
        self._start_worker(database_url, update_settings=True)

    @Slot()
    def _dispose_settings_window(self):
        """Deletes rejected export settings windows."""
        self._settings_window = None

    @Slot()
    def _update_settings_from_settings_window(self):
        """Pushes a new UpdateExporterSettingsCommand to the toolbox undo stack."""
        settings = self._settings_window.set_settings
        indexing_settings = self._settings_window.indexing_settings
        merging_settings = self._settings_window.merging_settings
        self._toolbox.undo_stack.push(
            UpdateSettings(
                self,
                settings,
                indexing_settings,
                merging_settings,
                self._settings_window.none_fallback,
                self._settings_window.none_export,
            )
        )

    def undo_or_redo_settings(self, settings, indexing_settings, merging_settings, none_fallback, none_export):
        """Updates the export settings for given database."""
        self._settings_pack.settings = settings
        self._settings_pack.indexing_settings = indexing_settings
        self._settings_pack.merging_settings = merging_settings
        self._settings_pack.none_fallback = none_fallback
        self._settings_pack.none_export = none_export
        if self._settings_window is not None:
            self._settings_window.reset_settings(
                self._settings_pack.settings,
                self._settings_pack.indexing_settings,
                self._settings_pack.merging_settings,
            )
        self._check_missing_specification()
        self._report_notifications()

    def item_dict(self):
        """Returns a dictionary corresponding to this item's configuration."""
        d = super().item_dict()
        databases = list()
        for db in self._database_model.items():
            db_dict = db.to_dict()
            serialized_url = serialize_url(db.url, self._project.project_dir)
            db_dict["database_url"] = serialized_url
            databases.append(db_dict)
        d["databases"] = databases
        d["settings_pack"] = self._settings_pack.to_dict()
        return d

    @staticmethod
    def from_dict(name, item_dict, toolbox, project):
        """See base class"""
        if "settings_pack" not in item_dict:
            return _legacy_from_dict(name, item_dict, toolbox, project)
        description, x, y = ProjectItem.parse_item_dict(item_dict)
        settings_dict = item_dict.get("settings_pack")
        if settings_dict is None:
            settings_pack = SettingsPack()
        else:
            try:
                settings_pack = SettingsPack.from_dict(settings_dict, toolbox)
            except gdx.GdxExportException as error:
                toolbox.msg_error.emit(f"Failed to fully restore GdxExporter settings: {error}")
                settings_pack = SettingsPack()
        databases = list()
        for db_dict in item_dict["databases"]:
            db = Database.from_dict(db_dict)
            db.url = clear_filter_configs(deserialize_path(db_dict["database_url"], project.project_dir))
            databases.append(db)
        output_time_stamps = item_dict.get("output_time_stamps", False)
        cancel_on_error = item_dict.get("cancel_on_error", True)
        return GdxExporter(
            name, description, x, y, toolbox, project, settings_pack, databases, output_time_stamps, cancel_on_error
        )

    def notify_destination(self, source_item):
        """See base class."""
        if source_item.item_type() == "Data Store":
            self._logger.msg.emit(
                f"Link established. Data Store <b>{source_item.name}</b> will be "
                f"exported to a .gdx file by <b>{self.name}</b> when executing."
            )
        elif source_item.item_type() == "Data Transformer":
            self._logger.msg.emit(
                f"Link established. Data transformed by <b>{source_item.name}</b> will be "
                f"exported to a .gdx file by <b>{self.name}</b> when executing."
            )
        else:
            super().notify_destination(source_item)

    def tear_down(self):
        """See base class."""
        super().tear_down()
        self._database_model.deleteLater()
        if self._worker is not None:
            self._worker.thread.quit()
            self._worker.thread.wait()
            self._worker.deleteLater()

    @staticmethod
    def upgrade_v1_to_v2(item_name, item_dict):
        """Upgrades item's dictionary from v1 to v2.

        Changes:
        - output_file_name and database_url stay the same but state is set to Fetching.

        Args:
            item_name (str): item's name
            item_dict (dict): Version 1 item dictionary

        Returns:
            dict: Version 2 item dictionary
        """
        old_settings_packs = item_dict.pop("settings_packs", list())
        new_settings_packs = list()
        for pack in old_settings_packs:
            new_pack = dict()
            new_pack["output_file_name"] = pack["output_file_name"]
            new_pack["state"] = SettingsState.FETCHING.value
            new_pack["database_url"] = pack["database_url"]
            new_settings_packs.append(new_pack)
        item_dict["settings_packs"] = new_settings_packs
        return item_dict


def _normalize_url(url):
    """
    Normalized url's path separators to their OS specific characters.

    This function is needed during the transition period from no-version to version 1 project files.
    It should be removed once we are using version 1 files.
    """
    return "sqlite:///" + url[10:].replace("/", os.sep)


def _legacy_from_dict(name, item_dict, toolbox, project):
    """
    Deserializes :class:`GdxExporter` from a legacy 0.5 item dict.

    Args:
        name (str): item's name
        item_dict (dict): serialized GdxExporter
        toolbox (ToolboxUI): Toolbox main widget
        project (SpineToolboxProject): project
    """
    description, x, y = ProjectItem.parse_item_dict(item_dict)
    settings_pack_dicts = item_dict.get("settings_packs")
    databases = list()
    if not settings_pack_dicts:
        settings_pack = SettingsPack()
    else:
        try:
            settings_pack = SettingsPack.from_dict(settings_pack_dicts[0], toolbox)
        except gdx.GdxExportException as error:
            toolbox.msg_error.emit(f"Failed to fully restore GdxExporter settings: {error}")
            settings_pack = SettingsPack()
        url_to_full_url = item_dict.get("urls")
        for pack in settings_pack_dicts:
            serialized_url = pack["database_url"]
            url = deserialize_path(serialized_url, project.project_dir)
            url = _normalize_url(url)
            db = Database()
            db.output_file_name = pack["output_file_name"]
            db.url = url if url_to_full_url is None else url_to_full_url[url]
            databases.append(db)
    cancel_on_error = item_dict.get("cancel_on_error", True)
    return GdxExporter(name, description, x, y, toolbox, project, settings_pack, databases, cancel_on_error)



######################################################################################################################
# Copyright (C) 2017-2021 Spine project consortium
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""
Contains Gimlet ExecutableItem class.

:author: P. Savolainen (VTT)
:date:   15.4.2020
"""

import os
import sys
import shutil
import uuid
from contextlib import ExitStack
from spine_engine.project_item.executable_item_base import ExecutableItemBase
from spine_engine.config import GIMLET_WORK_DIR_NAME
from spine_engine.utils.helpers import shorten
from spine_engine.utils.command_line_arguments import split_cmdline_args
from spine_engine.execution_managers.process_execution_manager import ProcessExecutionManager
from spine_engine.spine_engine import ItemExecutionFinishState
from spine_engine.project_item.project_item_resource import (
    CmdLineArg,
    cmd_line_arg_from_dict,
    expand_cmd_line_args,
    labelled_resource_args,
    labelled_resource_filepaths,
)
from .item_info import ItemInfo
from .utils import SHELLS


class ExecutableItem(ExecutableItemBase):
    def __init__(self, name, shell, cmd, work_dir, selected_files, project_dir, logger):
        """

        Args:
            name (str): Project item name
            shell (str): Shell name or empty string if no shell should be used
            cmd (list): Command to execute
            work_dir (str, optional): Full path to work directory
            selected_files (list): List of file paths that were selected
            project_dir (str): Absolute path to project directory
            logger (LoggerInterface): Logger instance
        """
        super().__init__(name, project_dir, logger)
        self.shell_name = shell
        self.cmd_list = cmd
        if work_dir is None:
            work_dir = os.path.join(self._data_dir, GIMLET_WORK_DIR_NAME)
        else:
            unique_dir_name = shorten(name) + "__" + uuid.uuid4().hex + "__toolbox"
            work_dir = os.path.join(work_dir, unique_dir_name)
        self._work_dir = work_dir
        self._resources = list()
        self._selected_files = selected_files
        self._exec_mngr = None

    @staticmethod
    def item_type():
        """Returns Gimlet's type identifier string."""
        return ItemInfo.item_type()

    @classmethod
    def from_dict(cls, item_dict, name, project_dir, app_settings, specifications, logger):
        """See base class."""
        if not item_dict["use_shell"]:
            shell = ""
        else:
            shell_index = item_dict["shell_index"]
            try:
                shell = SHELLS[shell_index]
            except IndexError:
                logger.msg_error.emit(f"Error: Unsupported shell_index in project item {name}")
                shell = ""
        cmd_list = [CmdLineArg(arg) for arg in split_cmdline_args(item_dict["cmd"])]
        cmd_list += [cmd_line_arg_from_dict(arg) for arg in item_dict["cmd_line_args"]]
        if item_dict["work_dir_mode"]:
            work_dir = None
        else:
            work_dir = app_settings.value("appSettings/workDir")
            if not work_dir:
                logger.msg_error.emit(f"Error: Work directory not set for project item {name}")
                work_dir = None
        selections = [label for label, selected in item_dict.get("file_selection", list()) if selected]
        return cls(name, shell, cmd_list, work_dir, selections, project_dir, logger)

    def stop_execution(self):
        """Stops executing this Gimlet."""
        super().stop_execution()
        if self._exec_mngr is not None:
            self._exec_mngr.stop_execution()
            self._exec_mngr = None

    def execute(self, forward_resources, backward_resources):
        """See base class."""
        if not super().execute(forward_resources, backward_resources):
            return ItemExecutionFinishState.FAILURE
        if not self._work_dir:
            self._logger.msg_warning.emit("Work directory not set.")
            return ItemExecutionFinishState.FAILURE
        if not self.cmd_list:
            self._logger.msg_warning.emit("No command to execute.")
            return ItemExecutionFinishState.FAILURE
        if sys.platform == "win32" and self.shell_name == "bash":
            self._logger.msg_warning.emit("Sorry, Bash shell is not supported on Windows.")
            return ItemExecutionFinishState.FAILURE
        if sys.platform != "win32" and (self.shell_name == "cmd.exe" or self.shell_name == "powershell.exe"):
            self._logger.msg_warning.emit(f"Sorry, selected shell is not supported on your platform [{sys.platform}]")
            return ItemExecutionFinishState.FAILURE
        with ExitStack() as stack:
            labelled_args = labelled_resource_args(forward_resources + backward_resources, stack)
            cmd_list = expand_cmd_line_args(self.cmd_list, labelled_args, self._logger)
        if not self.shell_name or self.shell_name == "bash":
            prgm = cmd_list.pop(0)
            self._exec_mngr = ProcessExecutionManager(self._logger, prgm, *cmd_list, workdir=self._work_dir)
        else:
            if self.shell_name == "cmd.exe":
                shell_prgm = "cmd.exe"
                cmd_list = ["/C"] + cmd_list
            elif self.shell_name == "powershell.exe":
                shell_prgm = "powershell.exe"
            else:
                self._logger.msg_error.emit(f"Unsupported shell: '{self.shell_name}'")
                return ItemExecutionFinishState.FAILURE
            self._exec_mngr = ProcessExecutionManager(self._logger, shell_prgm, *cmd_list, workdir=self._work_dir)
        # Copy selected files to work_dir
        selected_files = self._selected_files.copy()
        labelled_filepaths = labelled_resource_filepaths(forward_resources + backward_resources)
        for k, label in enumerate(selected_files):
            filepath = labelled_filepaths.get(label)
            if filepath is not None:
                selected_files[k] = filepath
        if not self._copy_files(selected_files, self._work_dir):
            return ItemExecutionFinishState.FAILURE
        # Make work directory anchor with path as tooltip
        work_anchor = (
            "<a style='color:#99CCFF;' title='"
            + self._work_dir
            + "' href='file:///"
            + self._work_dir
            + "'>work directory</a>"
        )
        self._logger.msg.emit(f"*** Executing in <b>{work_anchor}</b> ***")
        ret = self._exec_mngr.run_until_complete()
        if ret != 0:
            self._logger.msg_error.emit(f"Return code:{ret}")
        finish_state = ItemExecutionFinishState.SUCCESS if ret == 0 else ItemExecutionFinishState.FAILURE
        # Copy predecessor's resources so they can be passed to Gimlet's successors
        self._resources = forward_resources.copy()
        self._exec_mngr = None
        return finish_state

    def _output_resources_forward(self):
        """Returns output resources for forward execution.

        Returns:
            (list) List of ProjectItemResources.
        """
        return self._resources

    def _copy_files(self, files, work_dir):
        """Copies selected resources (files) to work directory.

        Args:
            files (list): List of full paths to files that will be copied to work dir
            work_dir (str): Full path to selected work dir

        Returns:
            bool: True when files were copied successfully, False when something went wrong
        """
        try:
            # Creates work_dir if it does not exist. Note: work_dir will be empty if len(files)==0.
            os.makedirs(work_dir, exist_ok=True)
        except OSError:
            self._logger.msg_error.emit(f"Creating directory <b>{work_dir}</b> failed")
            return False
        n_copied_files = 0
        for f in files:
            name = os.path.basename(f)
            dst_file = os.path.abspath(os.path.join(work_dir, name))
            try:
                # Copy file
                shutil.copyfile(f, dst_file)
                n_copied_files += 1
            except OSError:
                self._logger.msg_error.emit(f"\tCopying file <b>{f}</b> to <b>{dst_file}</b> failed")
                return False
        if n_copied_files == 0:
            self._logger.msg_warning.emit("\tNote: No files were copied")
        else:
            self._logger.msg.emit(f"\tCopied <b>{n_copied_files}</b> file(s)")
        return True



######################################################################################################################
# Copyright (C) 2017-2021 Spine project consortium
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""
Parameter merging settings window.

:author: A. Soininen (VTT)
:date:   19.2.2020
"""

from PySide2.QtCore import Qt, Signal, Slot
from PySide2.QtWidgets import QMessageBox, QWidget
from spinedb_api import DatabaseMapping
from .merging_error_flag import MergingErrorFlag
from .parameter_merging_settings import ParameterMergingSettings


class ParameterMergingSettingsWindow(QWidget):
    """A window which shows a list of ParameterMergingSettings widgets, one for each merged parameter."""

    settings_approved = Signal()
    """Emitted when the settings have been approved."""
    settings_rejected = Signal()
    """Emitted when the settings have been rejected."""

    def __init__(self, merging_settings, set_settings, database_path, parent):
        """
        Args:
            merging_settings (dict): a map from merged parameter name to a list of merging settings
            set_settings (SetSettings): set settings
            database_path (str): database URL
            parent (QWidget): a parent widget
        """
        from ..ui.parameter_merging_settings_window import Ui_Form  # pylint: disable=import-outside-toplevel

        super().__init__(parent, f=Qt.Window)
        self._merging_settings = merging_settings
        self._set_settings = set_settings
        self._entity_class_infos = None
        self._database_url = database_path
        self._ui = Ui_Form()
        self._ui.setupUi(self)
        self.setWindowTitle(f"Gdx Parameter Merging Settings")
        self._setting_widgets = list()
        for parameter_name, setting_list in merging_settings.items():
            for setting in setting_list:
                self._add_setting(parameter_name, setting)
        self._ui.button_box.accepted.connect(self._collect_and_hide)
        self._ui.button_box.rejected.connect(self._reject_and_close)
        self._ui.add_button.setEnabled(bool(database_path))
        self._ui.add_button.clicked.connect(self._add_empty_setting)

    @property
    def merging_settings(self):
        """a dict that maps merged parameter names to their merging settings"""
        return self._merging_settings

    def update(self):
        """Updates the settings according to changes in the database."""
        db_map = DatabaseMapping(self._database_url)
        try:
            self._entity_class_infos = _gather_entity_class_infos(db_map)
        finally:
            db_map.connection.close()
        for settings_widget in self._setting_widgets:
            settings_widget.update(self._entity_class_infos)

    def _add_setting(self, parameter_name=None, merging_setting=None):
        """Inserts a new settings widget to the widget list."""
        if self._entity_class_infos is None:
            if self._database_url:
                db_map = DatabaseMapping(self._database_url)
                try:
                    self._entity_class_infos = _gather_entity_class_infos(db_map)
                finally:
                    db_map.connection.close()
            else:
                self._entity_class_infos = _rebuild_entity_class_infos(self._merging_settings)
        settings_widget = ParameterMergingSettings(
            self._entity_class_infos, self._set_settings, self, parameter_name, merging_setting
        )
        settings_widget.removal_requested.connect(self._remove_setting)
        self._ui.settings_area_layout.insertWidget(0, settings_widget)
        self._setting_widgets.append(settings_widget)

    def _ok_to_accept(self):
        """Returns True if it is OK to accept the settings, otherwise shows a warning dialog and returns False."""
        for settings_widget in self._setting_widgets:
            flags = settings_widget.error_flags
            if flags & MergingErrorFlag.PARAMETER_NAME_MISSING:
                self._ui.setting_area.ensureWidgetVisible(settings_widget)
                message = "Parameter name is missing."
                QMessageBox.warning(self, "Parameter Name Missing", message)
                return False
            if flags & MergingErrorFlag.DOMAIN_NAME_MISSING:
                self._ui.setting_area.ensureWidgetVisible(settings_widget)
                message = "Domain name is missing."
                QMessageBox.warning(self, "Domain Name Missing", message)
                return False
            if flags & MergingErrorFlag.NO_PARAMETER_SELECTED:
                self._ui.setting_area.ensureWidgetVisible(settings_widget)
                message = "No domain selected for parameter merging."
                QMessageBox.warning(self, "Domain Selection Missing", message)
                return False
            if flags & MergingErrorFlag.DOMAIN_NAME_CLASH:
                self._ui.setting_area.ensureWidgetVisible(settings_widget)
                message = "A set with the same name already exists."
                QMessageBox.warning(self, "Invalid Domain Name", message)
                return False
        return True

    @Slot(bool)
    def _add_empty_setting(self, _):
        """Adds an empty settings widget to the widget list."""
        self._add_setting()

    @Slot(object)
    def _remove_setting(self, settings_widget):
        """Removes a setting widget from the widget list."""
        self._setting_widgets.remove(settings_widget)
        self._ui.settings_area_layout.removeWidget(settings_widget)
        settings_widget.deleteLater()

    @Slot()
    def _collect_and_hide(self):
        """Collects settings from individual ParameterMergingSettings widgets and hides the window."""
        if not self._ok_to_accept():
            return
        settings = dict()
        for widget in self._setting_widgets:
            setting_list = settings.setdefault(widget.parameter_name, list())
            setting_list.append(widget.merging_setting())
        self._merging_settings = settings
        self.settings_approved.emit()
        self.hide()

    @Slot()
    def _reject_and_close(self):
        """Emits settings_rejected and closes the window."""
        self.settings_rejected.emit()
        self.close()


class EntityClassInfo:
    """
    Contains information of an object_class or a relationship_class for use in the parameter merging widget.

    Attributes:
        name (str): entity class' name
        domain_names (tuple of str): object classes that index the entities in this class; for object classes this list
            contains the entity's name only, for relationship classes the list contains the relationship's object
            classes
        parameter_names (list of str): entity's defined parameters
    """

    def __init__(self, name, domain_names, parameter_names):
        """
        Args:
            name (str): entity class' name
            domain_names (tuple of str): object classes that index the entities in this class; for object classes it
                contains the entity's name only, for relationship classes it contains the relationship's object
                classes
            parameter_names (list of str): entity's defined parameters
        """
        self.name = name
        self.domain_names = domain_names
        self.parameter_names = parameter_names


def _gather_entity_class_infos(db_map):
    """
    Collects entity_class infos from database.

    Args:
        db_map (spinedb_api.DatabaseMapping): a database map
    Returns:
        list: a list of EntityClassInfo objects
    """
    infos = list()
    for object_class in db_map.query(db_map.object_class_sq):
        class_id = object_class.id
        parameter_definitions = db_map.query(db_map.parameter_definition_sq).filter_by(object_class_id=class_id)
        parameter_names = [definition.name for definition in parameter_definitions]
        infos.append(EntityClassInfo(object_class.name, (object_class.name,), parameter_names))
    for relationship_class in db_map.query(db_map.wide_relationship_class_sq):
        class_id = relationship_class.id
        parameter_definitions = db_map.query(db_map.parameter_definition_sq).filter_by(relationship_class_id=class_id)
        parameter_names = [definition.name for definition in parameter_definitions]
        domain_names = tuple(relationship_class.object_class_name_list.split(","))
        infos.append(EntityClassInfo(relationship_class.name, domain_names, parameter_names))
    return infos


def _rebuild_entity_class_infos(merging_settings):
    """
    Rebuilds entity class infos from given parameter merging settings.

    Args:
        merging_settings (dict): merging settings

    Returns:
        list of EntityClassInfo: entity class infos
    """
    infos = list()
    for settings_list in merging_settings.values():
        for merging_setting in settings_list:
            infos.append(
                EntityClassInfo(
                    merging_setting.previous_set, merging_setting.previous_domain_names, merging_setting.parameter_names
                )
            )
    return infos



######################################################################################################################
# Copyright (C) 2017-2021 Spine project consortium
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""
Gimlet class module.

:author: P. Savolainen (VTT)
:date:   15.4.2020
"""
import os
from PySide2.QtCore import QModelIndex, Slot, Qt
from spinetoolbox.project_item.project_item import ProjectItem
from spine_engine.config import GIMLET_WORK_DIR_NAME
from spine_engine.project_item.project_item_resource import cmd_line_arg_from_dict, LabelArg
from .item_info import ItemInfo
from .executable_item import ExecutableItem
from .commands import UpdateShellCheckBoxCommand, UpdateShellComboboxCommand, UpdatecmdCommand, UpdateWorkDirModeCommand
from ..commands import ChangeItemSelectionCommand, UpdateCmdLineArgsCommand
from ..models import GimletCommandLineArgsModel, CheckableFileListModel


class Gimlet(ProjectItem):
    """Gimlet class."""

    def __init__(
        self,
        name,
        description,
        x,
        y,
        toolbox,
        project,
        use_shell=True,
        shell_index=0,
        cmd="",
        cmd_line_args=None,
        selections=None,
        work_dir_mode=True,
    ):
        """
        Args:
            name (str): Project item name
            description (str): Description
            x (float): Initial X coordinate of item icon
            y (float): Initial Y coordinate of item icon
            toolbox (ToolboxUI): QMainWindow instance
            project (SpineToolboxProject): Project this item belongs to
            use_shell (bool): Use shell flag
            shell_index (int): Selected shell as index
            cmd (str): Command that this Gimlet executes at run time
            cmd_line_args (list, optional): Gimlet command line arguments
            selections (dict, optional): A mapping from file label to boolean 'checked' flag
            work_dir_mode (bool): True uses Gimlet's default work dir, False uses a unique work dir on every execution
        """
        super().__init__(name, description, x, y, project)
        self._toolbox = toolbox
        self._file_model = CheckableFileListModel(header_label="Available resources", draggable=True)
        self._resources_from_downstream = list()
        self._resources_from_upstream = list()
        self.use_shell = use_shell
        self.shell_index = shell_index
        self.cmd = cmd
        if cmd_line_args is None:
            cmd_line_args = []
        self.cmd_line_args = cmd_line_args
        self._cmdline_args_model = GimletCommandLineArgsModel(self)
        self._cmdline_args_model.args_updated.connect(self._push_update_cmd_line_args_command)
        self._populate_cmdline_args_model()
        self._file_model.set_initial_state(selections if selections is not None else dict())
        self._file_model.checked_state_changed.connect(self._push_file_selection_change_to_undo_stack)
        self._work_dir_mode = None
        self.update_work_dir_mode(work_dir_mode)
        self.default_gimlet_work_dir = os.path.join(self.data_dir, GIMLET_WORK_DIR_NAME)

    @staticmethod
    def item_type():
        """See base class."""
        return ItemInfo.item_type()

    @staticmethod
    def item_category():
        """See base class."""
        return ItemInfo.item_category()

    @property
    def executable_class(self):
        return ExecutableItem

    def make_signal_handler_dict(self):
        """Returns a dictionary of all shared signals and their handlers.
        This is to enable simpler connecting and disconnecting."""
        s = super().make_signal_handler_dict()
        s[self._properties_ui.checkBox_shell.stateChanged] = self.shell_checkbox_clicked
        s[self._properties_ui.comboBox_shell.activated] = self.shell_combobox_index_changed
        s[self._properties_ui.lineEdit_cmd.editingFinished] = self.cmd_edited
        s[self._properties_ui.radioButton_default.toggled] = self.push_work_dir_mode_cmd
        s[self._properties_ui.toolButton_add_file_path_arg.clicked] = self._add_selected_file_path_args
        s[self._properties_ui.toolButton_remove_arg.clicked] = self._remove_arg
        return s

    def restore_selections(self):
        """Restores selections into shared widgets when this project item is selected."""
        if not self._active:
            return
        self._properties_ui.treeView_cmdline_args.setModel(self._cmdline_args_model)
        self._properties_ui.treeView_files.setModel(self._file_model)
        self._properties_ui.checkBox_shell.setChecked(self.use_shell)
        if self.use_shell:
            self._properties_ui.comboBox_shell.setEnabled(True)
        else:
            self._properties_ui.comboBox_shell.setEnabled(False)
        self._properties_ui.comboBox_shell.setCurrentIndex(self.shell_index)
        self._properties_ui.lineEdit_cmd.setText(self.cmd)
        self.update_work_dir_button_state()

    def save_selections(self):
        """Saves selections in shared widgets for this project item into instance variables."""
        self._properties_ui.treeView_files.setModel(None)

    @Slot(int)
    def shell_checkbox_clicked(self, state):
        """Pushes a new shell check box command to undo stack.
        Pushing the command calls the commands redo method.

        Args:
            state (int): New check box state (Qt.CheckState enum)
        """
        use_shell = state == Qt.Checked
        if self.use_shell == use_shell:
            return
        self._toolbox.undo_stack.push(UpdateShellCheckBoxCommand(self, use_shell))

    def toggle_shell_state(self, use_shell):
        """Sets the use shell check box state. Disables shell
        combobox when shell check box is unchecked.

        Args:
            use_shell (bool): New check box state
        """
        self.use_shell = use_shell
        if not self._active:
            return
        # This does not trigger the stateChanged signal.
        self._properties_ui.checkBox_shell.setCheckState(Qt.Checked if use_shell else Qt.Unchecked)
        self._properties_ui.comboBox_shell.setEnabled(bool(use_shell))

    @Slot(int)
    def shell_combobox_index_changed(self, ind):
        """Pushes a shell combobox selection changed command to
        undo stack, which in turn calls set_shell_combobox_index() below.

        Args:
            ind (int): New index in combo box
        """
        if self.shell_index == ind:
            return
        self._toolbox.undo_stack.push(UpdateShellComboboxCommand(self, ind))

    def set_shell_combobox_index(self, ind):
        """Sets new index to shell combobox.

        Args:
            ind (int): New index in shell combo box
        """
        self.shell_index = ind
        if not self._active:
            return
        self._properties_ui.comboBox_shell.setCurrentIndex(ind)

    @Slot()
    def cmd_edited(self):
        """Updates the command instance variable when user has
        finished editing text in the line edit."""
        cmd = self._properties_ui.lineEdit_cmd.text().strip()
        if self.cmd == cmd:
            return
        self._toolbox.undo_stack.push(UpdatecmdCommand(self, cmd))

    def set_command(self, txt):
        self.cmd = txt
        if not self._active:
            return
        self._properties_ui.lineEdit_cmd.setText(txt)

    @Slot(bool)
    def _remove_arg(self, _=False):
        removed_rows = [index.row() for index in self._properties_ui.treeView_cmdline_args.selectedIndexes()]
        cmd_line_args = [arg for row, arg in enumerate(self.cmd_line_args) if row not in removed_rows]
        self._push_update_cmd_line_args_command(cmd_line_args)

    @Slot(bool)
    def _add_selected_file_path_args(self, _=False):
        new_args = [LabelArg(index.data()) for index in self._properties_ui.treeView_files.selectedIndexes()]
        self._push_update_cmd_line_args_command(self.cmd_line_args + new_args)

    @Slot(list)
    def _push_update_cmd_line_args_command(self, cmd_line_args):
        if self.cmd_line_args == cmd_line_args:
            return
        self._toolbox.undo_stack.push(UpdateCmdLineArgsCommand(self, cmd_line_args))

    def update_cmd_line_args(self, cmd_line_args):
        """Updates instance cmd line args list and sets the list as text to the line edit widget.

        Args:
            cmd_line_args (list): Tool cmd line args
        """
        self.cmd_line_args = cmd_line_args
        self._populate_cmdline_args_model()
        self._check_notifications()

    def _populate_cmdline_args_model(self):
        self._cmdline_args_model.reset_model(self.cmd_line_args)
        if self._active:
            self._properties_ui.treeView_cmdline_args.setFocus()

    @Slot(QModelIndex, bool)
    def _push_file_selection_change_to_undo_stack(self, index, checked):
        """Makes changes to file selection undoable."""
        self._toolbox.undo_stack.push(ChangeItemSelectionCommand(self.name, self._file_model, index, checked))

    @Slot(bool)
    def push_work_dir_mode_cmd(self, checked):
        """Pushes a new UpdateWorkDirModeCommand to the undo stack."""
        self._toolbox.undo_stack.push(UpdateWorkDirModeCommand(self, checked))

    def update_work_dir_mode(self, work_dir_mode):
        """Updates work_dir_mode setting.

        Args:
            work_dir_mode (bool): If True, work dir is set to this Gimlet's data dir,
            IF False, a unique work dir is created for every execution.
        """
        if self._work_dir_mode == work_dir_mode:
            return
        self._work_dir_mode = work_dir_mode
        self.update_work_dir_button_state()

    def update_work_dir_button_state(self):
        """Sets the work dir radio button check state according to
        work_dir_mode instance variable."""
        if not self._active:
            return
        self._properties_ui.radioButton_default.blockSignals(True)
        if self._work_dir_mode:
            self._properties_ui.radioButton_default.setChecked(True)
        else:
            self._properties_ui.radioButton_unique.setChecked(True)
        self._properties_ui.radioButton_default.blockSignals(False)

    def upstream_resources_updated(self, resources):
        self._resources_from_upstream = list(resources)
        self._update_files_and_cmd_line_args()

    def replace_resources_from_upstream(self, old, new):
        """See base class."""
        self._replace_resources(old, new, self._resources_from_upstream)

    def downstream_resources_updated(self, resources):
        self._resources_from_downstream = list(resources)
        self._update_files_and_cmd_line_args()

    def replace_resources_from_downstream(self, old, new):
        """See base class."""
        self._replace_resources(old, new, self._resources_from_downstream)

    def _replace_resources(self, old_resources, new_resources, resource_list):
        """Replaces resources.

        Modifies ``resource_list`` in-place!

        Args:
            old_resources (list of ProjectItemResource): old resource
            new_resources (list of ProjectItemResource): new resource
            resource_list (list of ProjectItemResource): current downstream or upstream resources
        """
        for old, new in zip(old_resources, new_resources):
            for i, resource in enumerate(resource_list):
                if resource == old:
                    resource_list[i] = new
                    break
            for i, arg in enumerate(self.cmd_line_args):
                if arg.arg == old.label:
                    self.cmd_line_args[i] = LabelArg(new.label)
                    break
        self.update_cmd_line_args(self.cmd_line_args)
        self._file_model.update(self._resources_from_upstream + self._resources_from_downstream)
        self._check_notifications()

    def _update_files_and_cmd_line_args(self):
        """Updates the file model and command line arguments with regards to available resources."""
        all_resources = self._resources_from_upstream + self._resources_from_downstream
        self._file_model.update(all_resources)
        self._check_notifications()
        update_args = list()
        resource_labels = {resource.label for resource in all_resources}
        for arg in self.cmd_line_args:
            if isinstance(arg, LabelArg):
                arg.missing = arg.arg not in resource_labels
            update_args.append(arg)
        self.update_cmd_line_args(update_args)

    def item_dict(self):
        """Returns a dictionary corresponding to this item."""
        d = super().item_dict()
        d["use_shell"] = self.use_shell
        d["shell_index"] = self.shell_index
        d["cmd"] = self.cmd
        selections = list()
        for row in range(self._file_model.rowCount()):
            label, selected = self._file_model.checked_data(self._file_model.index(row, 0))
            selections.append([label, selected])
        d["file_selection"] = selections
        d["work_dir_mode"] = self._work_dir_mode
        d["cmd_line_args"] = [arg.to_dict() for arg in self.cmd_line_args]
        return d

    @staticmethod
    def from_dict(name, item_dict, toolbox, project):
        """See base class."""
        description, x, y = ProjectItem.parse_item_dict(item_dict)
        use_shell = item_dict.get("use_shell", True)
        shel_index = item_dict.get("shell_index", 0)
        cmd = item_dict.get("cmd", "")
        selections = {label: selected for label, selected in item_dict.get("file_selection", list())}
        work_dir_mode = item_dict.get("work_dir_mode", True)
        cmd_line_args = item_dict.get("cmd_line_args", [])
        cmd_line_args = [cmd_line_arg_from_dict(arg) for arg in cmd_line_args]
        return Gimlet(
            name,
            description,
            x,
            y,
            toolbox,
            project,
            use_shell,
            shel_index,
            cmd,
            cmd_line_args,
            selections,
            work_dir_mode,
        )

    def notify_destination(self, source_item):
        """See base class."""
        if source_item.item_type() == "Data Connection":
            self._logger.msg.emit(
                f"Link established. Files from <b>{source_item.name}</b> are now available in <b>{self.name}</b>."
            )
            return
        if source_item.item_type() in ["Data Store", "Data Transformer", "Data Connection", "Tool", "Gimlet"]:
            self._logger.msg.emit("Link established")
            return
        if source_item.item_type() in ("GdxExporter", "Exporter"):
            self._logger.msg.emit(
                f"Link established. Files exported by <b>{source_item.name}</b> are "
                f"now available in <b>{self.name}</b>."
            )
            return
        super().notify_destination(source_item)

    def _check_notifications(self):
        """See base class."""
        self.clear_notifications()
        self.add_notification(
            "Gimlet is deprecated and will be removed in the future. Please replace this "
            "item with a <b>Tool</b> configured to run system commands."
        )

    def resources_for_direct_successors(self):
        """Returns resources for direct successors.

        This enables communication of resources between
        project items in the app.

        Returns:
            list: List of ProjectItemResources
        """
        return self._resources_from_upstream + self._resources_from_downstream



######################################################################################################################
# Copyright (C) 2017-2021 Spine project consortium
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""
A small widget to set up a database in Exporter properties tab.

:author: A. Soininen (VTT)
:date:   10.9.2019
"""

from PySide2.QtCore import Signal, Slot
from PySide2.QtWidgets import QWidget


class ExportListItem(QWidget):
    """A widget with few controls to select the output file name and open a settings window."""

    file_name_changed = Signal(str, str)
    """Emitted when the file name field is changed."""

    def __init__(self, url, file_name, parent=None):
        """
        Args:
            url (str): database URL
            file_name (str): relative path to the exported file name
            parent (QWidget): a parent widget
        """
        from ..ui.export_list_item import Ui_Form  # pylint: disable=import-outside-toplevel

        super().__init__(parent)
        self._ui = Ui_Form()
        self._url = url
        self._file_name = file_name
        self._ui.setupUi(self)
        self._ui.url_field.setText(url)
        self._ui.out_file_name_edit.setText(file_name)
        self._ui.out_file_name_edit.editingFinished.connect(self._emit_file_name_changed)

    @property
    def out_file_name_edit(self):
        """export file name QLineEdit"""
        return self._ui.out_file_name_edit

    @property
    def url_field(self):
        """Text in the database URL field."""
        return self._ui.url_field

    @Slot()
    def _emit_file_name_changed(self):
        """Emits file_name_changed signal."""
        file_name = self._ui.out_file_name_edit.text()
        if self._file_name == file_name:
            return
        self._file_name = file_name
        self.file_name_changed.emit(file_name, self._url)



######################################################################################################################
# Copyright (C) 2017-2021 Spine project consortium
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""
Contains :class:`RecordListModel`

:author: A. Soininen (VTT)
:date:   25.8.2020
"""
from PySide2.QtCore import QAbstractListModel, QModelIndex, Qt
from spinedb_api.spine_io.exporters import gdx
from ..list_utils import move_list_elements


class RecordListModel(QAbstractListModel):
    """A model to manage record ordering within domains and sets."""

    def __init__(self):
        super().__init__()
        self._records = gdx.LiteralRecords([])
        self._set_name = ""

    def data(self, index, role=Qt.DisplayRole):
        """With `role == Qt.DisplayRole` returns the record's keys as comma separated string."""
        if not index.isValid():
            return None
        if role == Qt.DisplayRole:
            keys = self._records.records[index.row()]
            return ", ".join(keys)
        return None

    def flags(self, index):
        if self._records.is_shufflable():
            return Qt.ItemIsEnabled | Qt.ItemIsSelectable
        return Qt.NoItemFlags

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        """Returns row and column header data."""
        if orientation == Qt.Horizontal:
            return ""
        return section + 1

    def moveRows(self, sourceParent, sourceRow, count, destinationParent, destinationChild):
        """
        Moves the records around.

        Args:
            sourceParent (QModelIndex): parent from which the rows are moved
            sourceRow (int): index of the first row to be moved
            count (int): number of rows to move
            destinationParent (QModelIndex): parent to which the rows are moved
            destinationChild (int): index where to insert the moved rows

        Returns:
            True if the operation was successful, False otherwise
        """
        if not self._records.is_shufflable():
            return False
        row_count = self.rowCount()
        if destinationChild < 0 or destinationChild >= row_count:
            return False
        last_source_row = sourceRow + count - 1
        row_after = destinationChild if sourceRow > destinationChild else destinationChild + 1
        self.beginMoveRows(sourceParent, sourceRow, last_source_row, destinationParent, row_after)
        self._records.shuffle(move_list_elements(self._records.records, sourceRow, last_source_row, destinationChild))
        self.endMoveRows()
        return True

    def reset(self, records, set_name):
        """Resets the model's record data."""
        self._set_name = set_name
        self.beginResetModel()
        self._records = records
        self.endResetModel()

    def rowCount(self, parent=QModelIndex()):
        """Returns the number of records in the model."""
        return len(self._records)

    def sort_alphabetically(self):
        """Sorts the record alphabetically"""
        if not self._records.is_shufflable():
            return
        self._records.shuffle(sorted(self._records.records))
        top_left = self.index(0, 0)
        bottom_right = self.index(len(self._records) - 1, 0)
        self.dataChanged.emit(top_left, bottom_right, [Qt.DisplayRole])



######################################################################################################################
# Copyright (C) 2017-2021 Spine project consortium
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""
Contains common & shared (Q)widgets.

:authors: M. Marin (KTH), P. Savolainen (VTT)
:date:   10.11.2020
"""

import os
from PySide2.QtCore import Qt, Signal, QUrl, QMimeData, Property
from PySide2.QtWidgets import QApplication, QTreeView, QStyledItemDelegate, QWidget, QDialog, QStatusBar
from PySide2.QtGui import QDrag, QIntValidator
from spinetoolbox.helpers import get_open_file_name_in_last_dir
from spinetoolbox.config import APPLICATION_PATH, STATUSBAR_SS
from .utils import convert_to_sqlalchemy_url


class ArgsTreeView(QTreeView):
    def dragEnterEvent(self, event):
        super().dragEnterEvent(event)
        event.accept()


class ReferencesTreeView(QTreeView):
    """Custom QTreeView class for Data Connection and View properties.

    Attributes:
        parent (QWidget): The parent of this view
    """

    files_dropped = Signal(list)
    del_key_pressed = Signal()

    def __init__(self, parent):
        """Initializes the view."""
        super().__init__(parent=parent)

    def dragEnterEvent(self, event):
        """Accepts file drops from the filesystem."""
        urls = event.mimeData().urls()
        for url in urls:
            if not url.isLocalFile():
                event.ignore()
                return
            if not os.path.isfile(url.toLocalFile()):
                event.ignore()
                return
        event.accept()
        event.setDropAction(Qt.LinkAction)

    def dragMoveEvent(self, event):
        """Accepts event."""
        event.accept()

    def dropEvent(self, event):
        """Emits files_dropped signal with a list of files for each dropped url."""
        self.files_dropped.emit([url.toLocalFile() for url in event.mimeData().urls()])

    def keyPressEvent(self, event):
        """Overridden method to make the view support deleting items with a delete key."""
        super().keyPressEvent(event)
        if event.key() == Qt.Key_Delete:
            self.del_key_pressed.emit()


class DataTreeView(QTreeView):
    """Custom QTreeView class for the 'Data' files in DataConnection properties.

    Attributes:
        parent (QWidget): The parent of this view
    """

    files_dropped = Signal(list)
    del_key_pressed = Signal()

    def __init__(self, parent):
        """Initializes the view."""
        super().__init__(parent=parent)
        self.drag_start_pos = None
        self.drag_indexes = list()

    def dragEnterEvent(self, event):
        """Accepts file drops from the filesystem."""
        urls = event.mimeData().urls()
        for url in urls:
            if not url.isLocalFile():
                event.ignore()
                return
            if not os.path.isfile(url.toLocalFile()):
                event.ignore()
                return
        event.accept()
        event.setDropAction(Qt.CopyAction)

    def dragMoveEvent(self, event):
        """Accepts event."""
        event.accept()

    def dropEvent(self, event):
        """Emits files_dropped signal with a list of files for each dropped url."""
        self.files_dropped.emit([url.toLocalFile() for url in event.mimeData().urls()])

    def mousePressEvent(self, event):
        """Registers drag start position."""
        if event.button() == Qt.LeftButton:
            self.drag_start_pos = event.pos()
            self.drag_indexes = self.selectedIndexes()
        super().mousePressEvent(event)

    def mouseMoveEvent(self, event):
        """Starts dragging action if needed."""
        if not event.buttons() & Qt.LeftButton:
            return
        if not self.drag_start_pos:
            return
        if not self.drag_indexes:
            return
        if (event.pos() - self.drag_start_pos).manhattanLength() < QApplication.startDragDistance():
            return
        drag = QDrag(self)
        mimeData = QMimeData()
        urls = list()
        for index in self.drag_indexes:
            file_path = index.data(Qt.UserRole)
            urls.append(QUrl.fromLocalFile(file_path))
        mimeData.setUrls(urls)
        drag.setMimeData(mimeData)
        icon = self.drag_indexes[0].data(Qt.DecorationRole)
        if icon:
            pixmap = icon.pixmap(32, 32)
            drag.setPixmap(pixmap)
            drag.setHotSpot(pixmap.rect().center())
        drag.exec_()

    def mouseReleaseEvent(self, event):
        """Forgets drag start position"""
        self.drag_start_pos = None
        super().mouseReleaseEvent(event)  # Fixes bug in extended selection

    def keyPressEvent(self, event):
        """Overridden method to make the view support deleting items with a delete key."""
        super().keyPressEvent(event)
        if event.key() == Qt.Key_Delete:
            self.del_key_pressed.emit()


class FilterEditDelegateBase(QStyledItemDelegate):
    """Base class for regexp filter edit delegates.

    Derived classes should reimplement at least ``createEditor()``.
    """

    def updateEditorGeometry(self, editor, option, index):
        top_left = option.rect.topLeft()
        popup_position = editor.parent().mapToGlobal(top_left)
        size_hint = editor.sizeHint()
        editor.setGeometry(
            popup_position.x(), popup_position.y(), max(option.rect.width(), size_hint.width()), size_hint.height()
        )


class FilterEdit(QWidget):
    """Filter regular expression editor."""

    def __init__(self, ui_form, parent):
        """
        Args:
            ui_form (Any): an interface from created from a .ui file
            parent (QWidget):
        """
        super().__init__(parent)
        self.setWindowFlags(Qt.Popup)
        self._ui = ui_form
        self._ui.setupUi(self)
        self._focused = False

    def focusInEvent(self, e):
        self._ui.regexp_line_edit.setFocus()
        self._ui.regexp_line_edit.selectAll()
        super().focusInEvent(e)

    def keyPressEvent(self, event):
        # Relay key press events to the regexp line edit. Otherwise we may lose the first letter.
        return self._ui.regexp_line_edit.keyPressEvent(event)

    def regexp(self):
        """Returns the current regular expression.

        Returns:
            str: regular expression
        """
        return self._ui.regexp_line_edit.text()

    def set_regexp(self, regexp):
        """Sets a regular expression for editing.

        Args:
            regexp (str): new regular expression
        """
        self._ui.regexp_line_edit.setText(regexp)

    regexp = Property(str, regexp, set_regexp, user=True)
    """Property used to communicate with the editor delegate."""


class UrlSelectorMixin:
    def _setup(self, dialects):
        self.ui.comboBox_dialect.addItems(dialects)
        self.ui.comboBox_dialect.setCurrentIndex(-1)
        self.ui.lineEdit_port.setValidator(QIntValidator())

    def enable_dialect(self, dialect):
        """Enable the given dialect in the item controls."""
        if dialect == "":
            self.enable_no_dialect()
        elif dialect == "sqlite":
            self.enable_sqlite()
        elif dialect == "mssql":
            import pyodbc  # pylint: disable=import-outside-toplevel

            dsns = pyodbc.dataSources()
            # Collect dsns which use the msodbcsql driver
            mssql_dsns = list()
            for key, value in dsns.items():
                if "msodbcsql" in value.lower():
                    mssql_dsns.append(key)
            if mssql_dsns:
                self.ui.comboBox_dsn.clear()
                self.ui.comboBox_dsn.addItems(mssql_dsns)
                self.ui.comboBox_dsn.setCurrentIndex(-1)
                self.enable_mssql()
            else:
                msg = "Please create a SQL Server ODBC Data Source first."
                self._logger.msg_warning.emit(msg)
        else:
            self.enable_common()

    def enable_no_dialect(self):
        """Adjust widget enabled status to default when no dialect is selected."""
        self.ui.comboBox_dialect.setEnabled(True)
        self.ui.comboBox_dsn.setEnabled(False)
        self.ui.toolButton_select_sqlite_file.setEnabled(False)
        self.ui.lineEdit_host.setEnabled(False)
        self.ui.lineEdit_port.setEnabled(False)
        self.ui.lineEdit_database.setEnabled(False)
        self.ui.lineEdit_username.setEnabled(False)
        self.ui.lineEdit_password.setEnabled(False)

    def enable_mssql(self):
        """Adjust controls to mssql connection specification."""
        self.ui.comboBox_dsn.setEnabled(True)
        self.ui.toolButton_select_sqlite_file.setEnabled(False)
        self.ui.lineEdit_host.setEnabled(False)
        self.ui.lineEdit_port.setEnabled(False)
        self.ui.lineEdit_database.setEnabled(False)
        self.ui.lineEdit_username.setEnabled(True)
        self.ui.lineEdit_password.setEnabled(True)
        self.ui.lineEdit_host.clear()
        self.ui.lineEdit_port.clear()
        self.ui.lineEdit_database.clear()

    def enable_sqlite(self):
        """Adjust controls to sqlite connection specification."""
        self.ui.comboBox_dsn.setEnabled(False)
        self.ui.comboBox_dsn.setCurrentIndex(-1)
        self.ui.toolButton_select_sqlite_file.setEnabled(True)
        self.ui.lineEdit_host.setEnabled(False)
        self.ui.lineEdit_port.setEnabled(False)
        self.ui.lineEdit_database.setEnabled(True)
        self.ui.lineEdit_username.setEnabled(False)
        self.ui.lineEdit_password.setEnabled(False)
        self.ui.lineEdit_host.clear()
        self.ui.lineEdit_port.clear()
        self.ui.lineEdit_username.clear()
        self.ui.lineEdit_password.clear()

    def enable_common(self):
        """Adjust controls to 'common' connection specification."""
        self.ui.comboBox_dsn.setEnabled(False)
        self.ui.comboBox_dsn.setCurrentIndex(-1)
        self.ui.toolButton_select_sqlite_file.setEnabled(False)
        self.ui.lineEdit_host.setEnabled(True)
        self.ui.lineEdit_port.setEnabled(True)
        self.ui.lineEdit_database.setEnabled(True)
        self.ui.lineEdit_username.setEnabled(True)
        self.ui.lineEdit_password.setEnabled(True)


class UrlSelector(UrlSelectorMixin, QDialog):
    msg_error = Signal(str)

    def __init__(self, toolbox, parent=None):
        from .ui.url_selector_widget import Ui_Dialog  # pylint: disable=import-outside-toplevel

        super().__init__(parent if parent is not None else toolbox)
        self._toolbox = toolbox
        self.ui = Ui_Dialog()
        self.ui.setupUi(self)
        # Add status bar to form
        self.statusbar = QStatusBar(self)
        self.statusbar.setFixedHeight(20)
        self.statusbar.setSizeGripEnabled(False)
        self.statusbar.setStyleSheet(STATUSBAR_SS)
        self.ui.horizontalLayout_statusbar_ph.addWidget(self.statusbar)
        self._sa_url = None
        self.ui.buttonBox.button(self.ui.buttonBox.Ok).setEnabled(False)
        self._setup(("mysql", "sqlite", "mssql", "postgresql", "oracle"))  # Others?
        self.ui.comboBox_dialect.activated[str].connect(self.enable_dialect)
        self.ui.comboBox_dialect.activated.connect(self._refresh_url)
        self.ui.toolButton_select_sqlite_file.clicked.connect(self._browse_sqlite_file)
        self.ui.lineEdit_username.editingFinished.connect(self._refresh_url)
        self.ui.lineEdit_password.editingFinished.connect(self._refresh_url)
        self.ui.lineEdit_host.editingFinished.connect(self._refresh_url)
        self.ui.lineEdit_port.editingFinished.connect(self._refresh_url)
        self.ui.lineEdit_database.editingFinished.connect(self._refresh_url)
        self.msg_error.connect(self.statusbar.showMessage)

    @property
    def url(self):
        if self._sa_url is None:
            return ""
        return str(self._sa_url)

    def _refresh_url(self):
        url = {
            "dialect": self.ui.comboBox_dialect.currentText(),
            "host": self.ui.lineEdit_host.text(),
            "port": self.ui.lineEdit_port.text(),
            "database": self.ui.lineEdit_database.text(),
            "username": self.ui.lineEdit_username.text(),
            "password": self.ui.lineEdit_password.text(),
        }
        self._sa_url = convert_to_sqlalchemy_url(url, logger=self)
        self.ui.buttonBox.button(self.ui.buttonBox.Ok).setEnabled(self._sa_url is not None)

    def _browse_sqlite_file(self):
        filter_ = "*.sqlite;;*.*"
        key = "selectImportSourceSQLiteFile"
        filepath, _ = get_open_file_name_in_last_dir(
            self._toolbox.qsettings(), key, self, "Select a SQLite file", APPLICATION_PATH, filter_=filter_
        )
        if not filepath:
            return
        self.ui.lineEdit_database.setText(filepath)
        self._refresh_url()



######################################################################################################################
# Copyright (C) 2017-2021 Spine project consortium
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""
Contains list helper functions for list manipulation.

:author: A. Soininen (VTT)
:date:   12.12.2019
"""

from PySide2.QtCore import QModelIndex


def move_list_elements(originals, first, last, target):
    """
    Moves elements in a list.

    Args:
        originals (list): a list
        first (int): index of the first element to move
        last (int): index of the last element to move
        target (int): index where the elements `[first:last]` should be inserted

    Return:
        a new list with the elements moved
    """
    trashable = list(originals)
    elements_to_move = list(originals[first : last + 1])
    del trashable[first : last + 1]
    elements_that_come_before = trashable[:target]
    elements_that_come_after = trashable[target:]
    brave_new_list = elements_that_come_before + elements_to_move + elements_that_come_after
    return brave_new_list


def move_selected_elements_by(list_view, delta):
    """
    Moves selected items in a QListView by given delta.

    Args:
        list_view (QListView): a list view
        delta (int): positive values move the items up, negative down
    """
    selection_model = list_view.selectionModel()
    selected_rows = sorted(selection_model.selectedRows())
    if not selected_rows:
        return
    first_row = selected_rows[0].row()
    contiguous_selections = [[first_row, 1]]
    current_contiguous_chunk = contiguous_selections[0]
    for row in selected_rows[1:]:
        if row == current_contiguous_chunk[0] + 1:
            current_contiguous_chunk[1] += 1
        else:
            contiguous_selections.append((row, 1))
            current_contiguous_chunk = contiguous_selections[-1]
    model = list_view.model()
    for chunk in contiguous_selections:
        model.moveRows(QModelIndex(), chunk[0], chunk[1], QModelIndex(), chunk[0] + delta)



######################################################################################################################
# Copyright (C) 2017-2021 Spine project consortium
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""
Contains :class:`ExporterNotifications`.

:author: A. Soininen (VTT)
:date:   11.1.2022
"""


class ExporterNotifications:
    """
    Holds flags for different exporter error conditions.

    Attributes:
        duplicate_output_file_name (bool): True if there are duplicate output file names
        missing_output_file_name (bool): True if the output file name is missing
        missing_specification (bool): True if export specification is missing
    """

    def __init__(self):
        self.duplicate_output_file_name = False
        self.missing_output_file_name = False
        self.missing_specification = False



######################################################################################################################
# Copyright (C) 2017-2021 Spine project consortium
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""
Undo/redo commands for the Gimlet project item.

:authors: P. Savolainen (VTT)
:date:   30.4.2020
"""
from spinetoolbox.project_commands import SpineToolboxCommand


class UpdateShellCheckBoxCommand(SpineToolboxCommand):
    """Command to update Gimlet shell check box state."""

    def __init__(self, gimlet, use_shell):
        """

        Args:
            gimlet (spine_items.gimlet.gimlet.Gimlet): The Gimlet issuing the command
            use_shell (bool): New check box state
        """
        super().__init__()
        self._gimlet = gimlet
        self._redo_use_shell = use_shell
        self._undo_use_shell = not use_shell
        self.setText(f"change {gimlet.name} shell check box state")

    def redo(self):
        self._gimlet.toggle_shell_state(self._redo_use_shell)

    def undo(self):
        self._gimlet.toggle_shell_state(self._undo_use_shell)


class UpdateShellComboboxCommand(SpineToolboxCommand):
    """Command to update Gimlet shell combobox state."""

    def __init__(self, gimlet, new_index):
        """

        Args:
            gimlet (spine_items.gimlet.gimlet.Gimlet): The Gimlet issuing the command
            new_index (int): New combobox index
        """
        super().__init__()
        self._gimlet = gimlet
        self._redo_combobox_index = new_index
        self._undo_combobox_index = self._gimlet.shell_index  # Previous index
        self.setText(f"change {gimlet.name} shell combobox selection")

    def redo(self):
        self._gimlet.set_shell_combobox_index(self._redo_combobox_index)

    def undo(self):
        self._gimlet.set_shell_combobox_index(self._undo_combobox_index)


class UpdatecmdCommand(SpineToolboxCommand):
    """Command to update Gimlet command line edit."""

    def __init__(self, gimlet, txt):
        """

        Args:
            gimlet (spine_items.gimlet.gimlet.Gimlet): The Gimlet issuing the command
            txt (str): New text in command line edit after editing is finished
        """
        super().__init__()
        self._gimlet = gimlet
        self._redo_cmd = txt
        self._undo_cmd = self._gimlet.cmd
        self.setText(f"change command in {gimlet.name}")

    def redo(self):
        self._gimlet.set_command(self._redo_cmd)

    def undo(self):
        self._gimlet.set_command(self._undo_cmd)


class UpdateWorkDirModeCommand(SpineToolboxCommand):
    def __init__(self, gimlet, work_dir_mode):
        """Command to update Gimlet work_in_dir setting.

        Args:
            gimlet (Gimlet): The Gimlet
            work_dir_mode (bool): True or False
        """
        super().__init__()
        self._gimlet = gimlet
        self._work_dir_mode = work_dir_mode
        self.setText(f"change work dir mode setting of {gimlet.name}")

    def redo(self):
        self._gimlet.update_work_dir_mode(self._work_dir_mode)

    def undo(self):
        self._gimlet.update_work_dir_mode(not self._work_dir_mode)
